from netket.hilbert.spin import Spin
from netket.sampler.rules import MetropolisRule


class GlobalSpinFlipRule(MetropolisRule):
    r"""A Metropolis rule that flips all spins in a configuration."""

    def transition(rule, sampler, machine, parameters, state, key, σ):
        if not isinstance(sampler.hilbert, Spin):
            raise TypeError(
                "GlobalSpinFlipRule is only compatible with Spin hilbert spaces"
            )
        return -σ, None



from collections.abc import Callable
from functools import partial

import jax
import jax.numpy as jnp
from jax.sharding import NamedSharding, PartitionSpec as P

from netket import jax as nkjax
from netket.utils.types import Array

from netket import config

from netket._src.ngd.kwargs import ensure_accepts_kwargs


@partial(
    jax.jit,
    static_argnames=(
        "solver_fn",
        "mode",
    ),
)
def _compute_sr_update(
    O_L,
    dv,
    *,
    diag_shift: float | Array,
    solver_fn: Callable[[Array, Array], Array],
    mode: str,
    proj_reg: float | Array | None = None,
    momentum: float | Array | None = None,
    old_updates: Array | None = None,
    params_structure,
):
    # We concretize the solver function to ensure it accepts the additional argument `dv`.
    # Typically solvers only accept the matrix and the right-hand side.
    solver_fn = ensure_accepts_kwargs(solver_fn, "dv")

    if proj_reg is not None:
        raise ValueError("proj_reg not implemented for SR")

    # (np, #ns) x (#ns) -> (np) - where the sum over #ns is done automatically
    F = O_L.T @ dv

    # Add momentum term to F: F + λ μ dθ_{i-1}
    if momentum is not None:
        F = F + diag_shift * momentum * old_updates

    # This does the contraction (np, #ns) x (#ns, np) -> (np, np).
    matrix = O_L.T @ O_L
    matrix_side = matrix.shape[-1]  # * it can be ns or 2*ns, depending on mode

    shifted_matrix = jax.lax.add(
        matrix, diag_shift * jnp.eye(matrix_side, dtype=matrix.dtype)
    )
    updates = solver_fn(shifted_matrix, F, dv=dv)

    # Some solvers return a tuple, some others do not.
    if isinstance(updates, tuple):
        updates, info = updates
        if info is None:
            info = {}
    else:
        info = {}

    if momentum is not None:
        old_updates = updates

    # If complex mode and we have complex parameters, we need
    # To repack the real coefficients in order to get complex updates
    if mode == "complex" and nkjax.tree_leaf_iscomplex(params_structure):
        num_p = updates.shape[-1] // 2
        updates = updates[:num_p] + 1j * updates[num_p:]

    if config.netket_experimental_sharding:
        updates = jax.lax.with_sharding_constraint(
            updates, NamedSharding(jax.sharding.get_abstract_mesh(), P())
        )

    return updates, old_updates, info



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Documentation-specific citation utilities for NetKet.

This module provides functions for generating formatted citations for documentation.
"""

import re
from netket._version import __version__
from .core import get_all_citations, register_citation, _citation_registry


def _extract_doi_from_bibtex(bibtex_string: str) -> str | None:
    """
    Extract DOI from a BibTeX entry.

    Args:
        bibtex_string: BibTeX entry as string

    Returns:
        DOI string if found, None otherwise
    """
    # Look for doi field in various formats
    doi_patterns = [
        r"doi\s*=\s*\{([^}]+)\}",
        r'doi\s*=\s*"([^"]+)"',
        r"doi\s*=\s*([^,\s}]+)",
    ]

    for pattern in doi_patterns:
        match = re.search(pattern, bibtex_string, re.IGNORECASE)
        if match:
            return match.group(1).strip()

    return None


def _extract_url_from_bibtex(bibtex_string: str) -> str | None:
    """
    Extract URL from a BibTeX entry.

    Args:
        bibtex_string: BibTeX entry as string

    Returns:
        URL string if found, None otherwise
    """
    # Look for url field in various formats
    url_patterns = [
        r"url\s*=\s*\{([^}]+)\}",
        r'url\s*=\s*"([^"]+)"',
        r"url\s*=\s*([^,\s}]+)",
    ]

    for pattern in url_patterns:
        match = re.search(pattern, bibtex_string, re.IGNORECASE)
        if match:
            return match.group(1).strip()

    return None


def _extract_title_from_bibtex(bibtex_string: str) -> str | None:
    """
    Extract title from a BibTeX entry.

    Args:
        bibtex_string: BibTeX entry as string

    Returns:
        Title string if found, None otherwise
    """
    # Look for title field in various formats
    title_patterns = [
        r"title\s*=\s*\{([^}]+)\}",
        r'title\s*=\s*"([^"]+)"',
    ]

    for pattern in title_patterns:
        match = re.search(pattern, bibtex_string, re.IGNORECASE)
        if match:
            # Remove LaTeX formatting
            title = match.group(1).strip()
            title = re.sub(r"\{([^}]*)\}", r"\1", title)  # Remove braces
            return title

    return None


def get_citations_for_docs(format_type: str = "detailed") -> str:
    """
    Generate HTML formatted citations for documentation.

    Args:
        format_type: Either 'detailed' or 'simple'

    Returns:
        HTML string with formatted citations
    """
    # NetKet is already imported at this point since this function is called from within NetKet

    netket_version_parts = __version__.split(".")[:2]
    netket_version = ".".join(netket_version_parts)

    # Register NetKet's own references if not already present
    netket_papers = ["netket3:2021", "netket2:2019"]
    if not any(paper_key in _citation_registry for paper_key in netket_papers):
        register_citation(
            netket_papers, message=f"This work used NetKet {netket_version}"
        )

    all_citations = get_all_citations()
    if not all_citations:
        return '<div class="admonition note"><p class="admonition-title">Note</p><p>No citations registered.</p></div>'

    # Group citations by their group_id
    citation_groups = {}
    for key, info in all_citations.items():
        group_id = info.get("group_id", key)
        if group_id not in citation_groups:
            citation_groups[group_id] = {
                "keys": [],
                "condition": info.get("condition", "").strip(),
                "message": info.get("message", "").strip(),
                "bibtex_entries": [],
                "titles_and_links": [],
            }
        citation_groups[group_id]["keys"].append(key)
        if "bibtex" in info:
            bibtex = info["bibtex"]
            citation_groups[group_id]["bibtex_entries"].append(bibtex)

            # Extract title, DOI, and URL for linking
            title = _extract_title_from_bibtex(bibtex)
            doi = _extract_doi_from_bibtex(bibtex)
            url = _extract_url_from_bibtex(bibtex)

            # Determine link URL (prioritize DOI over URL)
            link_url = None
            if doi:
                link_url = f"https://doi.org/{doi}"
            elif url:
                link_url = url

            citation_groups[group_id]["titles_and_links"].append(
                {"key": key, "title": title, "link_url": link_url}
            )

    html_parts = []
    html_parts.append('<div class="netket-citations">')
    html_parts.append("<h3>Registered Citations</h3>")

    if format_type == "detailed":
        html_parts.append(
            "<p>The following citations have been registered based on your NetKet usage:</p>"
        )

        for group_id, group_info in citation_groups.items():
            html_parts.append('<div class="citation-group">')

            # Show condition if present
            if group_info["condition"]:
                html_parts.append(
                    f'<p><strong>When:</strong> {group_info["condition"]}</p>'
                )

            # Show message if present
            if group_info["message"]:
                html_parts.append(
                    f'<p><strong>Citation message:</strong> {group_info["message"]}</p>'
                )

            # Show citation keys with linked titles if available
            if group_info["titles_and_links"]:
                refs_parts = []
                for entry in group_info["titles_and_links"]:
                    if entry["title"] and entry["link_url"]:
                        refs_parts.append(
                            f'<code>{entry["key"]}</code>: <a href="{entry["link_url"]}" target="_blank">{entry["title"]}</a>'
                        )
                    elif entry["title"]:
                        refs_parts.append(
                            f'<code>{entry["key"]}</code>: {entry["title"]}'
                        )
                    else:
                        refs_parts.append(f'<code>{entry["key"]}</code>')
                html_parts.append(
                    f'<p><strong>References:</strong> {"; ".join(refs_parts)}</p>'
                )
            else:
                keys_str = ", ".join(
                    f"<code>{key}</code>" for key in group_info["keys"]
                )
                html_parts.append(f"<p><strong>References:</strong> {keys_str}</p>")

            # Show BibTeX entries if available
            if group_info["bibtex_entries"]:
                html_parts.append("<details><summary>BibTeX entries</summary>")
                for bibtex in group_info["bibtex_entries"]:
                    html_parts.append(
                        f'<pre><code class="language-bibtex">{bibtex}</code></pre>'
                    )
                html_parts.append("</details>")

            html_parts.append("</div><hr/>")

    else:  # simple format
        html_parts.append("<ul>")
        for group_id, group_info in citation_groups.items():
            keys_str = ", ".join(group_info["keys"])
            condition_str = (
                f" ({group_info['condition']})" if group_info["condition"] else ""
            )
            html_parts.append(f"<li><code>{keys_str}</code>{condition_str}</li>")
        html_parts.append("</ul>")

    html_parts.append("</div>")

    return "\n".join(html_parts)



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

from netket.hilbert import DiscreteHilbert, Qubit, Spin, Fock, SpinOrbitalFermions
from netket.symmetry.group import PermutationGroup, PointGroup
from netket.graph.space_group import TranslationGroup, SpaceGroup

from netket._src.symmetry.representation import Representation
from netket._src.symmetry.representation_construction import (
    physical_to_logical_permutation_group,
)
from netket._src.operator.permutation.construct import (
    construct_permutation_operator,
)


def canonical_representation(
    hilbert: DiscreteHilbert, group: PermutationGroup, warn: bool = True
) -> Representation:
    r"""
    Construct the representation of a permutation group on a many-body
    Hilbert space where each permutation is mapped to the permutation operator
    that permutes the local operators.

    This function creates a representation of a permutation group that acts on the
    physical sites of a lattice or graph, extended to act on the full many-body
    Hilbert space. The representation operators depend on the Hilbert space type.

    More precisely, on a spin Hilbert space, the permutation operator associated
    to the permutation :math:`\sigma` is the operator :math:`P_\sigma` such that
    for any local operator :math:`A_k` acting on site :math:`k`, the permutation
    operator transforms it as :math:`P_\sigma A_k P_\sigma^\dagger = A_{\sigma(k)}`.
    For a fermionic Hilbert space, the operator :math:`P_\sigma` is defined such
    that for any site :math:`k`, the creation operator :math:`c_k^\dagger`
    transforms as :math:`P_\sigma c_k^\dagger P_\sigma^\dagger = c_{\sigma(k)}^\dagger`.
    The relevant permutation operators are therefore different depending on the
    type of Hilbert space. For more details, see :doc:`/advanced/symmetry`.

    .. note::
        This function is **experimental** and its API may change in future releases.

    .. warning::
        This function constructs the specific representation of a permutation
        group described above and in the documentation. For other non-spatial
        symmetries, the appropriate representation can be constructed manually
        using the constructor of :class:`~netket.symmetry.Representation`.
        To safeguard against mistaken uses of this function, passing a
        :class:`~netket.symmetry.group.PermutationGroup` that is not also an
        instance of a subclass designed specifically to describe a subgroup of
        the space group, that is, :class:`~netket.graph.space_group.SpaceGroup`,
        :class:`~netket.symmetry.group.PointGroup`, or
        :class:`~netket.graph.space_group.TranslationGroup`, will raise a warning.

    Args:
        hilbert: The Hilbert space on which the representation acts. This can be any
            discrete Hilbert space, including spin systems (e.g., :class:`~netket.hilbert.Spin`,
            :class:`~netket.hilbert.Qubit`) or fermionic systems
            (:class:`~netket.hilbert.SpinOrbitalFermions`).
        group: A :class:`~netket.utils.group.PermutationGroup` object describing the
            symmetry group. This can be obtained from lattice methods like
            :meth:`~netket.graph.Lattice.translation_group`,
            :meth:`~netket.graph.Lattice.point_group`,
            :meth:`~netket.graph.Lattice.space_group`, etc., or be a custom permutation group.
        warn: If False, disable the warning on passing a
            :class:`~netket.symmetry.group.PermutationGroup` that is not a
            :class:`~netket.graph.space_group.SpaceGroup`,
            a :class:`~netket.symmetry.group.PointGroup`, or a
            :class:`~netket.graph.space_group.TranslationGroup`.

    Returns:
        A :class:`~netket.symmetry.Representation` object encoding the action of the
        permutation group on the Hilbert space. The representation includes a dictionary
        mapping each group element to the corresponding operator that acts on quantum states.

    Examples:
        Get the representation of the translation group on a spin system:

        >>> import netket as nk
        >>> lattice = nk.graph.Square(4)
        >>> hilbert = nk.hilbert.Spin(0.5, N=lattice.n_nodes)
        >>> trans_group = lattice.translation_group()
        >>> rep = nk.symmetry.canonical_representation(hilbert, trans_group)

        Get the representation of a partial symmetry (translations along x only):

        >>> trans_x = lattice.translation_group(dim=0)
        >>> rep_x = nk.symmetry.canonical_representation(hilbert, trans_x)

        Use a point group symmetry (e.g., C4 rotations):

        >>> pg = lattice.point_group(nk.symmetry.group.planar.C(4))
        >>> rep_c4 = nk.symmetry.canonical_representation(hilbert, pg)

        Use the full space group (translations + point group):

        >>> sg = lattice.space_group()
        >>> rep_sg = nk.symmetry.canonical_representation(hilbert, sg)

        This approach works with any :class:`~netket.utils.group.PermutationGroup`,
        allowing you to construct representations for specific subgroups of the full
        symmetry group or even use custom permutation groups, though in the latter
        case, care should be taken to only use this function when intending to
        construct the specific representation of that custom permutation group
        detailed above.

        For fermionic systems, the permutations are automatically extended to account
        for multiple spin species:

        >>> fermion_hilbert = nk.hilbert.SpinOrbitalFermions(
        ...     n_orbitals=lattice.n_nodes,
        ...     s=1/2,
        ...     n_fermions_per_spin=(4, 4)
        ... )
        >>> rep_fermion = nk.symmetry.canonical_representation(
        ...     fermion_hilbert, trans_group
        ... )

    See Also:
        - :class:`~netket.symmetry.Representation`: The representation class that is returned.
        - :class:`~netket.utils.group.PermutationGroup`: Permutation group class.
        - :meth:`~netket.graph.Lattice.translation_group`: Get the translation group of a lattice.
        - :meth:`~netket.graph.Lattice.space_group`: Get the space group of a lattice.
        - :meth:`~netket.graph.Lattice.point_group`: Get the point group of a lattice.
    """

    if not isinstance(group, PermutationGroup):
        raise NotImplementedError(
            "Only `PermutationGroup` are supported (for now,\n"
            "please open an issue to report what new groups we should support)."
        )

    if not isinstance(hilbert, (Qubit, Spin, Fock, SpinOrbitalFermions)):
        raise ValueError(
            "The permutation operators of this representation "
            "are only defined for a Hilbert space of the class Qubit, Spin, Fock, "
            "or SpinOrbitalFermions."
        )
    if isinstance(hilbert, (Qubit, Spin, Fock)) and not group.degree == hilbert.size:
        raise ValueError(
            "In the case of a spin Hilbert space, the permutations of group "
            "should be permutations over hilbert.size elements."
        )
    if (
        isinstance(hilbert, (SpinOrbitalFermions))
        and not group.degree == hilbert.n_orbitals
    ):
        raise ValueError(
            "In the case of a fermionic Hilbert space, the permutations of "
            "group should be permutations over hilbert.n_orbitals elements."
        )
    if warn and not isinstance(group, (SpaceGroup, TranslationGroup, PointGroup)):
        warnings.warn(
            "This function constructs a specific representation "
            "of the given permutation group that corresponds to spatial "
            "symmetries, as described above and in the documentation.\n\n"
            "Make sure that this is the intended representation. "
            "To disable this warning, pass a group of the class "
            "SpaceGroup, PointGroup, or TranslationGroup, "
            "or pass no_warning=True.\n\n"
        )

    # If fermionic hilbert space, increase the size of the permutation so it
    # matches the number of single-particle states.
    group = physical_to_logical_permutation_group(group, hilbert)
    return Representation(
        group, {perm: construct_permutation_operator(hilbert, perm) for perm in group}
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from netket.operator._discrete_operator import DiscreteOperator

from netket.operator._prod.base import ProductOperator


class ProductDiscreteOperator(ProductOperator, DiscreteOperator):
    def __init__(self, *operators, coefficient=1.0, dtype=None):
        if not all(isinstance(op, DiscreteOperator) for op in operators):
            raise TypeError(
                "Arguments to ProductDiscreteOperator must all be "
                "subtypes of DiscreteOperator. However the types are:\n\n"
                f"{list(type(op) for op in operators)}\n"
            )
        self._initialized = False
        super().__init__(
            operators, operators[0].hilbert, coefficient=coefficient, dtype=dtype
        )

    def _setup(self, force: bool = False):
        if not self._initialized:
            self._initialized = True

    def get_conn_flattened(
        self,
        x: np.ndarray,
        sections: np.ndarray,
        pad: bool = False,
    ) -> tuple[np.ndarray, np.ndarray]:
        raise NotImplementedError

    def to_jax_operator(self) -> "ProductDiscreteJaxOperator":  # noqa: F821
        """
        Returns the standard (numba) version of this operator, which is an
        instance of {class}`nk.operator.Ising`.
        """
        from netket.operator._prod.discrete_jax_operator import (
            ProductDiscreteJaxOperator,
        )

        ops_jax = tuple(op.to_jax_operator() for op in self.operators)
        return ProductDiscreteJaxOperator(
            *ops_jax, coefficient=self.coefficient, dtype=self.dtype
        )



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import jax
import jax.numpy as jnp
from jax.tree_util import register_pytree_node_class

from functools import partial

from netket.hilbert import SpinOrbitalFermions
from netket.utils.group import Permutation
from netket.utils.types import DType

from netket._src.operator.permutation.base import (
    PermutationOperatorBase,
)
from netket._src.operator.permutation.trace_utils import (
    get_subset_occupations,
    get_parity_sum,
)


def get_parity(array: jax.Array) -> jax.Array:
    """
    Count the parity of an array.
    This is the number of inversions in the array modulo 2.
    An inversion is a pair (i, j) such that i < j
    and array[i] > array[j].
    """
    batch_dims = array.shape[:-1]
    inversion_matrix = array[..., :, jnp.newaxis] > array[..., jnp.newaxis, :]
    upper_triangular_mask = jnp.triu(
        jnp.ones((*batch_dims, array.shape[-1], array.shape[-1]), dtype=bool), k=1
    )
    inversion_count = jnp.sum(inversion_matrix & upper_triangular_mask, axis=(-2, -1))
    return inversion_count % 2


def get_occupied_orbitals(x: jax.Array, n_fermions: int) -> jax.Array:
    """Return the indices of the occupied orbitals
    in a given SpinOrbitalFermions state n"""
    batch_dims, physical_dims = x.shape[:-1], x.shape[-1]
    x = x.reshape(-1, physical_dims)
    occupied_orbitals = _get_occupied_orbitals(x, n_fermions)
    return occupied_orbitals.reshape(*batch_dims, n_fermions)


@partial(jax.vmap, in_axes=(0, None))
def _get_occupied_orbitals(x: jax.Array, n_fermions: int) -> jax.Array:
    return x.nonzero(size=n_fermions)[0]


@partial(jax.jit, static_argnames=("n_fermions",))
def get_antisymmetric_signs(
    x: jax.Array, permutation: jax.Array, n_fermions: int
) -> jax.Array:
    """Return the sign of the permutation for a batch of fermionic Fock states x."""
    occupied = get_occupied_orbitals(x, n_fermions)
    permuted = permutation[occupied]
    parity = get_parity(permuted)
    sign = 1 - 2 * parity
    return sign


def check_permutation_compatibility(
    permutation: Permutation, partition: list, constraints: list
):
    """
    Check that a permutation is compatible with constraints.

    Given a permutation that acts on n elements and a partition of those n elements
    that constrains bitstrings to contain exactly constraints[k] bits 1 in the
    partition subset partition[k], check that the permutation only maps compatible
    bitstrings to compatible bitstrings. In other words, it checks that the
    permutation operator defined by that permutation is well-defined on the
    Hilbert space specified by the constraints.

    It can be shown that it is equivalent to check that each partition subset
    is mapped by the inverse permutation onto another partition subset that has
    the same constraint. This equivalence only holds if the constraints are
    non-trivial, that is, only if the bitstrings are not constrained to be either
    only 0 or only 1 within any partition subset.

    Args:
        permutation: The permutation whose compatibility we check.
        partition: The partition that defines the constraint. It must be a partition of `range(n)` specified as a
            list where the k-th element is a list that specifies the elements of the k-th partition subset.
        constraints: The occupation constraint on each partition subset.
            The k-th constraint must be an integer contained strictly between 0
            and the size of the k-th partition subset.
    """

    for partition_subset in partition:
        partition_subset.sort()

    index_set = sorted(sum(partition, start=[]))
    n = index_set[-1]

    assert index_set == list(range(n + 1)), "`partition` is not a proper partition."

    for partition_subset, constraint in zip(partition, constraints, strict=True):
        assert (
            0 < constraint < len(partition_subset)
        ), "There is a trivial or invalid constraint."

    is_compatible = True

    for index, partition_subset in enumerate(partition):

        partition_subset_preimage = sorted(
            permutation.inverse_permutation_array[partition_subset]
        )

        if partition_subset_preimage in partition:
            image_index = partition.index(partition_subset_preimage)
            if constraints[image_index] != constraints[index]:
                is_compatible = False
        else:
            is_compatible = False

    return is_compatible


@register_pytree_node_class
class PermutationOperatorFermion(PermutationOperatorBase):
    """
    Operators corresponding to the permutation of the orbitals belonging to a
    second quantized fermionic space. Used to express lattice symmetries of
    fermionic Hamiltonians and symmetrize variational states with respect to
    these symmetries.

    This permutation operator differs from
    :class:`netket.operator.permutation.PermutationOperator` due to the
    fermionic sign that is required to properly define permutations of fermionic
    orbitals.

    .. warning::

        Only works for spaces with a fixed number of fermions.

    For mathematical details on the definition of a permutation operator
    and its justification, we refer to :doc:`/advanced/symmetry`.
    """

    def __init__(
        self,
        hilbert: SpinOrbitalFermions,
        permutation: Permutation,
        dtype: DType | None = float,
    ):
        """
        Construct the permutation operator.

        Args:
            hilbert: The fermionic Hilbert space.
            permutation: The permutation that defines the operator as
                specified in :doc:`/advanced/symmetry`.
        """
        assert isinstance(hilbert, SpinOrbitalFermions)

        if hilbert.n_fermions is None:
            raise TypeError("The Hilbert space must have a fixed number of fermions.")

        if hilbert.n_fermions_per_spin[0] is not None:

            if (
                sum(
                    constraint == 0 or constraint == hilbert.n_orbitals
                    for constraint in hilbert.n_fermions_per_spin
                )
                > 0
            ):
                warnings.warn(
                    "The Hilbert space contains a spin sector that is "
                    "either empty or fully occupied. The compatibility of the "
                    "permutation could not be checked. The trivial spin sector "
                    "should be removed from the Hilbert space."
                )

            else:
                partition = list(
                    list(range(k * hilbert.n_orbitals, (k + 1) * hilbert.n_orbitals))
                    for k in range(hilbert.n_spin_subsectors)
                )
                is_compatible = check_permutation_compatibility(
                    permutation, partition, list(hilbert.n_fermions_per_spin)
                )
                assert (
                    is_compatible
                ), "The permutation is not compatible with the Hilbert space constraints"

        super().__init__(hilbert, permutation, dtype=dtype)

    def _get_signs(self, x):
        return get_antisymmetric_signs(
            x, self.permutation.inverse_permutation_array, self.hilbert.n_fermions
        )

    def get_conn_padded(self, x):
        r"""Finds the connected elements of the Operator.

        Starting from a batch of quantum numbers :math:`x={x_1, ... x_n}` of
        size :math:`B \times M` where :math:`B` size of the batch and :math:`M`
        size of the hilbert space, finds all states :math:`y_i^1, ..., y_i^K`
        connected to every :math:`x_i`.

        Returns a matrix of size :math:`B \times K_{max} \times M` where
        :math:`K_{max}` is the maximum number of connections for every
        :math:`y_i`.

        .. warning::

            Unlike most other operators defined in NetKet, a permutation operator
            is not Hermitian, and we thus have to be careful about the definition of
            connected elements. NetKet defines connected elements of :math:`x` as the
            configurations :math:`x'` such that :math:`\langle x | P_\sigma | x' \rangle` .
            Therefore, the connected elements are the configurations found in the
            image of :math:`x` by :math:`P_{\sigma^{-1}}` , and not :math:`P_\sigma` .

        Args:
            x : A N-tensor of shape :math:`(...,hilbert.size)` containing
                the batch/batches of quantum numbers :math:`x`.

        Returns:
            **(x_primes, mels)**: The connected states x', in a N+1-tensor and an
            N-tensor containing the matrix elements :math:`O(x,x')`
            associated to each x' for every batch.
        """

        x = jnp.asarray(x)
        connected_elements = x.at[..., None, self.permutation.permutation_array].get(
            unique_indices=True, mode="promise_in_bounds"
        )
        signs = self._get_signs(x).astype(self.dtype)
        return connected_elements, signs[..., jnp.newaxis]

    def trace(self) -> int:
        partition_labels = sum(
            [
                self.hilbert.n_orbitals * [k]
                for k in range(self.hilbert.n_spin_subsectors)
            ],
            start=[],
        )
        cycle_decomposition = self.permutation.cycle_decomposition()
        cycle_occupation = get_subset_occupations(partition_labels, cycle_decomposition)
        return get_parity_sum(cycle_occupation, self.hilbert.n_fermions_per_spin)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import platform

from ._common import version, is_available
from ._cpu_info import cpu_info

PLATFORM = sys.platform

STD_SPACE = 20


def printfmt(key, value=None, *, indent=0, pre="", alignment=STD_SPACE):
    INDENT_STR = "- "

    if indent > 0:
        indent_spaces = indent * 2
        alignment -= indent_spaces + len(INDENT_STR)

        pre = f"{pre}{'' : <{indent_spaces}}{INDENT_STR}"

    if value is not None:
        print(f"{pre}{key : <{alignment}} : {value}", flush=True)
    else:
        print(f"{pre}{key : <{alignment}}", flush=True)


def _fmt_device(dev):
    return f"<{dev.id}: {dev.device_kind}>"


def info():
    print("====================================================")
    print("==         NetKet Diagnostic Information          ==")
    print("====================================================")

    # try to import version without import netket itself
    from .. import _version

    printfmt("NetKet version", _version.version)
    print()

    print("# Python")
    printfmt("implementation", platform.python_implementation(), indent=1)
    printfmt("version", platform.python_version(), indent=1)
    printfmt("distribution", platform.python_compiler(), indent=1)
    printfmt("path", sys.executable, indent=1)
    print()

    # Try to detect platform
    print("# Host information")
    printfmt("System      ", platform.platform(), indent=1)
    printfmt("Architecture", platform.machine(), indent=1)

    # Try to query cpu info
    platform_info = cpu_info()

    printfmt("AVX", platform_info["supports_avx"], indent=1)
    printfmt("AVX2", platform_info["supports_avx2"], indent=1)
    printfmt("Cores", platform_info["core_count"], indent=1)
    print()

    # try to load jax
    print("# NetKet dependencies")
    printfmt("numpy", version("numpy"), indent=1)
    printfmt("jaxlib", version("jaxlib"), indent=1)
    printfmt("jax", version("jax"), indent=1)
    printfmt("equinox", version("equinox"), indent=1)
    printfmt("flax", version("flax"), indent=1)
    printfmt("optax", version("optax"), indent=1)
    printfmt("numba", version("numba"), indent=1)
    printfmt("netket", version("netket"), indent=1)
    print()

    if is_available("jax"):
        print("# Jax ")
        import jax

        backends = _jax_backends()
        printfmt("backends", backends, indent=1)
        for backend in backends:
            printfmt(
                f"{backend}",
                [_fmt_device(dev) for dev in jax.devices(backend)],
                indent=2,
            )
        print()


def _jax_backends():
    import jax

    backends = []
    for backend in ["cpu", "gpu", "tpu"]:
        try:
            jax.devices(backend)
        except RuntimeError:
            pass
        else:
            backends.append(backend)
    return backends


info()



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Internal utility functions to support jax sharding natively within netket.
All functions in here are not part of the public API, internal, and may change without warning.
"""

from netket.jax.sharding.debug import inspect as inspect

from netket.jax.sharding.fixed_sharding_decorator import (
    sharding_decorator as sharding_decorator,
    _increase_SHARD_MAP_STACK_LEVEL as _increase_SHARD_MAP_STACK_LEVEL,
    _get_SHARD_MAP_STACK_LEVEL as _get_SHARD_MAP_STACK_LEVEL,
)

from netket.jax.sharding.fixed_sharding_utils import (
    shard_along_axis as shard_along_axis,
    with_samples_sharding_constraint as with_samples_sharding_constraint,
    extract_replicated as extract_replicated,
    distribute_to_devices_along_axis as distribute_to_devices_along_axis,
    gather as gather,
)

from netket.jax.sharding.flexible_sharding import (
    pad_axis_for_sharding as pad_axis_for_sharding,
)


from netket.utils import _hide_submodules
from netket.utils.deprecation import warn_deprecation as _warn_deprecation


# TODO: Deprecated in July 2025, remove eventually. It was internal
def __getattr__(name):
    """Handle deprecated attribute access with warnings."""
    if name == "SHARD_MAP_STACK_LEVEL":
        _warn_deprecation(
            "Accessing netket.jax.sharding.SHARD_MAP_STACK_LEVEL directly is deprecated. Use _get_SHARD_MAP_STACK_LEVEL() instead.",
        )
        return _get_SHARD_MAP_STACK_LEVEL()
    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")


_hide_submodules(__name__)



import jax.numpy as jnp
import flax
import jax

from netket.experimental.observable import AbstractObservable

from netket.utils.types import DType
from netket.utils.numbers import is_scalar
from netket.vqs import VariationalState, FullSumState, MCState
from netket.operator import AbstractOperator
from netket.jax import HashablePartial


class InfidelityOperator(AbstractObservable):
    r"""
    Infidelity operator computing the infidelity between an input variational state :math:`|\Psi\rangle` and a target state :math:`|\Phi\rangle`.

    The target state can be defined in two ways:

    1. as a variational state that is passed as `target_state`.
    2. as a state obtained by applying an operator :math:`U` to a variational state :math:`|\Phi\rangle`, i.e., :math:`|\Phi\rangle \equiv U|\Phi\rangle`.

    The infidelity :math:`I` among two variational states :math:`|\Psi\rangle` and :math:`|\Phi\rangle` is defined as:

    .. math::

        I = 1 - \frac{|\langle\Psi|\Phi\rangle|^2}{\langle\Psi|\Psi\rangle \langle\Phi|\Phi\rangle} = 1 - \frac{\langle\Psi|\hat{I}_{op}|\Psi\rangle}{\langle\Psi|\Psi\rangle},

    where:

    .. math::

        \hat{I}_{op} = \frac{|\Phi\rangle\langle\Phi|}{\langle\Phi|\Phi\rangle}.

    The Monte Carlo estimator of :math:`I` is:

    .. math::

        I = \mathbb{E}_{\chi}[ I_{loc}(x,y) ] = \mathbb{E}_{\chi}\left[ \frac{\langle x|\Phi\rangle \langle y|\Psi\rangle}{\langle x|\Psi\rangle \langle y|\Phi\rangle} \right],

    where :math:`\chi(x, y) = \frac{|\Psi(x)|^2 |\Phi(y)|^2}{\langle\Psi|\Psi\rangle \langle\Phi|\Phi\rangle}` is the joint Born distribution. This estimator
    can be utilized both when :math:`|\Phi\rangle = |\Phi\rangle` and when :math:`|\Phi\rangle = U|\Phi\rangle`, with :math:`U` a (unitary or
    non-unitary) operator. We remark that sampling from :math:`U|\Phi\rangle` is more expensive than sampling from an autonomous state.

    For details see `Sinibaldi et al. <https://quantum-journal.org/papers/q-2023-10-10-1131/>`_ and `Gravina et al. <https://quantum-journal.org/papers/q-2025-07-22-1803/>`_.
    """

    def __init__(
        self,
        target_state: VariationalState,
        *,
        operator: AbstractOperator = None,
        cv_coeff: float | None = -0.5,
        dtype: DType | None = None,
    ):
        """
        Args:
            target_state: The target state :math:`|\\Phi\\rangle` against which to compute the infidelity.
                This can be any VariationalState (MCState, FullSumState, etc.).
            operator: Optional operator :math:`U` to be applied to the target state, such that :math:`|\\Phi\\rangle \\equiv U |\\Phi\\rangle`.
                If None, the target state is used directly. When provided, the infidelity is computed
                with respect to the transformed state :math:`U |\\Phi\\rangle`.
            cv_coeff: Optional control variate coefficient for variance reduction in Monte Carlo
                estimation (see `Sinibaldi et al. <https://quantum-journal.org/papers/q-2023-10-10-1131/>`_).
                If None, no control variate is used. Default to the optimal value -0.5.
            dtype: Data type for computations.
        """
        super().__init__(target_state.hilbert)

        if not isinstance(target_state, VariationalState):
            raise TypeError("The first argument should be a variational target.")

        if cv_coeff is not None:
            cv_coeff = jnp.array(cv_coeff)

            if (not is_scalar(cv_coeff)) or jnp.iscomplex(cv_coeff):
                raise TypeError("`cv_coeff` should be a real scalar number or None.")

            if isinstance(target_state, FullSumState):
                cv_coeff = None

        if operator is not None:

            def _logpsi_fun(apply_fun, variables, x, *args):
                variables_applyfun, O = flax.core.pop(variables, "operator")

                xp, mels = O.get_conn_padded(x)
                xp = xp.reshape(-1, x.shape[-1])
                logpsi_xp = apply_fun(variables_applyfun, xp, *args)
                logpsi_xp = logpsi_xp.reshape(mels.shape)

                return jax.scipy.special.logsumexp(
                    logpsi_xp.astype(complex), axis=-1, b=mels
                )

            logpsi_fun = HashablePartial(_logpsi_fun, target_state._apply_fun)

            if isinstance(target_state, MCState):
                self._target_state = MCState(
                    sampler=target_state.sampler,
                    apply_fun=logpsi_fun,
                    n_samples=target_state.n_samples,
                    variables=flax.core.copy(
                        target_state.variables, {"operator": operator}
                    ),
                )
            if isinstance(target_state, FullSumState):
                self._target_state = FullSumState(
                    hilbert=target_state.hilbert,
                    apply_fun=logpsi_fun,
                    variables=flax.core.copy(
                        target_state.variables, {"operator": operator}
                    ),
                )
        else:
            self._target_state = target_state

        self._cv_coeff = cv_coeff
        self._dtype = dtype

    @property
    def target_state(self):
        return self._target_state

    @property
    def cv_coeff(self):
        return self._cv_coeff

    def __repr__(self):
        return f"InfidelityOperator(target_state={self.target_state}, cv_coeff={self.cv_coeff})"



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable

from abc import ABC

import jax.numpy as jnp

from netket.hilbert import AbstractHilbert
from netket.utils.types import Array
from netket.jax import canonicalize_dtypes
from netket.utils.numbers import is_scalar

from netket.operator._abstract_operator import AbstractOperator
from netket.operator._discrete_operator import DiscreteOperator
from netket.operator._discrete_operator_jax import DiscreteJaxOperator

# from .._continuous_operator import ContinuousOperator


def _flatten_prodoperators(operators: Iterable[AbstractOperator], coefficient: float):
    """Flatten sumoperators inside of operators."""
    new_operators = []
    new_coeff = coefficient
    for op in operators:
        if isinstance(op, ProductOperator):
            new_operators.extend(op.operators)
            new_coeff = new_coeff * op.coefficient
        else:
            new_operators.append(op)
    return new_operators, new_coeff


class ProductOperator(ABC):
    """
    Base class for product of quantum operators.

    This class represents a product of quantum operators with a coefficient,
    implementing the mathematical concept of :math:`c \\prod_i \\hat{H}_i` where
    :math:`c` is a scalar coefficient and :math:`\\hat{H}_i` are quantum operators.

    The class uses a dispatch mechanism to automatically select the appropriate
    specialized subclass based on the types of operators being multiplied:

    * :class:`~netket.operator._prod.discrete_jax_operator.ProductDiscreteJaxOperator`
      for products of :class:`~netket.operator.DiscreteJaxOperator` instances
    * :class:`~netket.operator._prod.discrete_operator.ProductDiscreteOperator`
      for products of :class:`~netket.operator.DiscreteOperator` instances
    * :class:`~netket.operator._prod.operator.ProductGenericOperator`
      for mixed operator types or fallback cases.

    Products can be constructed using the ``@`` operator (matrix multiplication) or by
    directly instantiating this class with a list of operators.

    Examples:
        Creating a product of Pauli operators:

        >>> import netket as nk
        >>> hi = nk.hilbert.Spin(s=1/2, N=4)
        >>> sx0 = nk.operator.spin.sigmax(hi, 0)
        >>> sz1 = nk.operator.spin.sigmaz(hi, 1)
        >>> sy2 = nk.operator.spin.sigmay(hi, 2)
        >>> # Using ProductOperator constructor
        >>> product1 = nk.operator.ProductOperator(sx0, sz1, sy2)
        >>> print(product1)  # doctest: +SKIP
        ProductDiscreteJaxOperator with terms:
         ∙ 1.0
         ∙ PauliStringsJax(n_sites=4, hilbert=Spin(s=1/2, N=4), n_strings=1)
         ∙ PauliStringsJax(n_sites=4, hilbert=Spin(s=1/2, N=4), n_strings=1)
         ∙ PauliStringsJax(n_sites=4, hilbert=Spin(s=1/2, N=4), n_strings=1)

        Creating a product with explicit coefficient:

        >>> # Direct instantiation
        >>> product2 = nk.operator.ProductOperator(sx0, sz1, coefficient=2.0)
        >>> print(product2.coefficient)
        2.0

        Products of products are automatically flattened:

        >>> prod_a = nk.operator.ProductOperator(sx0, sz1)
        >>> prod_b = nk.operator.ProductOperator(sy2, nk.operator.spin.sigmaz(hi, 3))
        >>> combined = nk.operator.ProductOperator(prod_a, prod_b)
        >>> print(len(combined.operators))  # All 4 operators are flattened
        4

        Scaling a product:

        >>> scaled = 3.0 * product1
        >>> print(scaled.coefficient)
        (3+0j)
    """

    def __new__(cls, *args, **kwargs):
        # This logic overrides the constructor, such that if someone tries to
        # construct this class directly by calling `SumOperator(...)`
        # it will construct either a DiscreteHilbert or TensorDiscreteHilbert
        from netket.operator._prod.operator import ProductGenericOperator
        from netket.operator._prod.discrete_operator import ProductDiscreteOperator
        from netket.operator._prod.discrete_jax_operator import (
            ProductDiscreteJaxOperator,
        )

        # from .continuous import ProductContinuousOperator

        if cls is ProductOperator:
            if all(isinstance(op, DiscreteJaxOperator) for op in args):
                cls = ProductDiscreteJaxOperator
            elif all(isinstance(op, DiscreteOperator) for op in args):
                cls = ProductDiscreteOperator
            # elif all(isinstance(op, ContinuousOperator) for op in args):
            #    cls = ProductContinuousOperator
            else:
                cls = ProductGenericOperator
        return super().__new__(cls)

    def __init__(
        self,
        operators: Iterable[AbstractHilbert],
        *args,
        coefficient: float = 1.0,
        dtype=None,
        **kwargs,
    ):
        r"""Constructs a Product of Operators.

        Args:
            operators: An iterable of quantum operators to be multiplied. All operators
                must act on the same Hilbert space.
            coefficient: Scalar coefficient for the product. Default is 1.0.
            dtype: Data type for the coefficient. If None, it will be inferred
                from the operators and coefficient.
        """
        hi_spaces = [op.hilbert for op in operators]
        if not all(hi == hi_spaces[0] for hi in hi_spaces):
            raise NotImplementedError(
                "Cannot construct a ProductOperator for operators on different Hilbert Spaces"
            )

        if not is_scalar(coefficient):
            raise TypeError(f"Coefficient must be a scalar but was {coefficient}")

        operators, coefficient = _flatten_prodoperators(operators, coefficient)

        dtype = canonicalize_dtypes(float, *operators, coefficient, dtype=dtype)

        self._operators = tuple(operators)
        self._coefficient = jnp.asarray(coefficient, dtype=dtype)
        self._dtype = dtype

        super().__init__(
            *args, **kwargs
        )  # forwards all unused arguments so that this class is a mixin.

    @property
    def dtype(self):
        return self._dtype

    @property
    def operators(self) -> tuple[AbstractOperator, ...]:
        """The tuple of all operators in the terms of this product. All operators
        are multiplied together with the scalar coefficient.
        """
        return self._operators

    @property
    def coefficient(self) -> Array:
        """The scalar coefficient multiplying the product of operators."""
        return self._coefficient

    def __repr__(self) -> str:
        strs = [f"{type(self).__name__} with terms:"]
        # Format the coefficient
        coefficient_str = f" ∙ {self.coefficient}"
        strs.append(coefficient_str)

        # Format each operator
        for op in self.operators:
            op_str = str(op).splitlines()
            # Format the first line of the operator
            formatted_op = f" ∙ {op_str[0]}"
            # Add subsequent lines with proper indentation
            if len(op_str) > 1:
                formatted_op += "\n" + "\n".join(
                    f"{' ' * len(coefficient_str)}{line}" for line in op_str[1:]
                )
            strs.append(formatted_op)

        return "\n".join(strs)

    def __mul__(self, other):
        if is_scalar(other):
            return ProductOperator(
                *self.operators, coefficient=self.coefficient * other, dtype=self.dtype
            )
        return super().__mul__(other)

    def __matmul__(self, other):
        if not isinstance(other, AbstractOperator):
            return super().__matmul__(other)

        if isinstance(other, ProductOperator):
            ops = self.operators + other.operators
            coeff = self.coefficient * other.coefficient
            dtype = self.dtype if self.dtype == other.dtype else None
        else:
            ops = (*self.operators, other)
            coeff = self.coefficient
            dtype = self.dtype if self.dtype == other.dtype else None

        return ProductOperator(*ops, coefficient=coeff, dtype=dtype)

    def __rmatmul__(self, other):
        if not isinstance(other, AbstractOperator):
            return super().__rmatmul__(other)

        ops = (other, *self.operators)
        coeff = self.coefficient
        dtype = self.dtype if self.dtype == other.dtype else None

        return ProductOperator(*ops, coefficient=coeff, dtype=dtype)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Literal, overload

import jax
import jax.numpy as jnp
from jax.sharding import NamedSharding, PartitionSpec as P

from netket.utils import random_seed, config
from netket.utils.types import PRNGKeyT, SeedT


def PRNGKey(seed: SeedT | None = None, *, root: int = 0) -> PRNGKeyT:
    """
    Initialises a PRNGKey using an optional starting seed.

    If using sharding, the returned key will be replicated on every process.

    Args:
        seed: An optional integer value to use as seed
        root: the master rank, used when running with multiple nodes (default 0)

    Returns:
        A sharded/broadcasted :func:`jax.random.PRNGKey`.

    """
    if seed is None:
        seed = random_seed()

    if isinstance(seed, int):
        # We can't sync the PRNGKey, so we can only sinc integer seeds
        # see https://github.com/google/jax/pull/16511
        if config.netket_experimental_sharding and jax.process_count() > 1:  # type: ignore[attr-defined]
            # TODO: use stable jax function
            from jax.experimental import multihost_utils

            seed = int(
                multihost_utils.broadcast_one_to_all(
                    seed, is_source=jax.process_index() == root
                ).item()
            )

        key = jax.random.key(seed)
    elif isinstance(seed, jax.Array):
        if jnp.issubdtype(seed.dtype, jax.dtypes.prng_key):
            # new-style keys
            key = seed
        elif jnp.issubdtype(seed.dtype, jax.numpy.unsignedinteger):
            # old-style PRNGKey
            key = seed
        else:
            raise TypeError(
                f"Expected seed to be an integer or a PRNGKey, got {type(seed)}, {seed.dtype} : {seed}"
            )
    else:
        raise TypeError(f"unsupported type {type(seed)}")

    mesh = jax.sharding.get_abstract_mesh()
    if not mesh.empty:
        if len(mesh.explicit_axes) >= 0:
            key = jax.sharding.reshard(key, P())
        else:
            sharding = NamedSharding(mesh, P())
            key = jax.lax.with_sharding_constraint(key, sharding)
    return key


class PRNGSeq:
    """
    A sequence of PRNG keys generated based on an initial key.
    """

    def __init__(self, base_key: SeedT | None = None):
        if base_key is None:
            base_key = PRNGKey()
        elif isinstance(base_key, int):
            base_key = PRNGKey(base_key)
        self._current: PRNGKeyT = base_key

    def __iter__(self):
        return self

    def __next__(self):
        self._current = jax.random.split(self._current, num=1)[0]
        return self._current

    def next(self):
        return self.__next__()

    def take(self, num: int):
        """
        Returns an array of `num` PRNG keys and advances the iterator accordingly.
        """
        keys = jax.random.split(self._current, num=num + 1)
        self._current = keys[-1]
        return keys[:-1]


def _bcast_key(key, root=0) -> PRNGKeyT:
    """
    Utility function equivalent to broadcast a random key to multiple jax processes
    and make sure it is the same everywhere.
    """
    is_new_style_key = jnp.issubdtype(key.dtype, jax.dtypes.prng_key)

    if is_new_style_key:
        _impl = jax.random.key_impl(key)
        key = jax.random.key_data(key)

    if is_new_style_key:
        key = jax.random.wrap_key_data(key, impl=_impl)  # type: ignore[arg-type]

    return key


@overload
def batch_choice(
    key: PRNGKeyT, a: jax.Array, p: jax.Array, *, return_prob: Literal[False] = False
) -> jax.Array: ...


@overload
def batch_choice(
    key: PRNGKeyT, a: jax.Array, p: jax.Array, *, return_prob: Literal[True]
) -> tuple[jax.Array, jax.Array]: ...


def batch_choice(
    key: PRNGKeyT, a: jax.Array, p: jax.Array, *, return_prob: bool = False
) -> jax.Array | tuple[jax.Array, jax.Array]:
    """
    Batched version of `jax.random.choice`.

    Args:
      key: a PRNGKey used as the random key.
      a: 1D array. Random samples are generated from its elements.
      p: 2D array of shape `(batch_size, a.size)`. Each slice `p[i, :]` is
        the probabilities associated with entries in `a` to generate a sample
        at the index `i` of the output. Can be unnormalized.
      return_prob: If True, also return the probabilities of the chosen samples.

    Returns:
      The generated samples as an 1D array of shape `(batch_size,)`.
      If return_prob is True, returns a tuple of (samples, probabilities).
    """
    p_cumsum = p.cumsum(axis=1)
    r = p_cumsum[:, -1:] * jax.random.uniform(key, shape=(p.shape[0], 1))
    indices = (r > p_cumsum).sum(axis=1)
    out = a[indices]

    if return_prob:
        probs = jnp.take_along_axis(p, indices[:, None], axis=1).squeeze(axis=1)
        return out, probs
    else:
        return out



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Core citation system implementation for NetKet.

This module provides the internal implementation for citation tracking and bibliography generation.
"""

import re
from pathlib import Path

from netket._version import __version__

from rich.console import Console
from rich.text import Text

# Global citation registry
_citation_registry: dict[str, dict[str, str]] = {}


def _parse_bibtex_entry(bibtex_string: str) -> dict[str, str]:
    """
    Parse a simple BibTeX entry to extract key information.

    Args:
        bibtex_string: BibTeX entry as string

    Returns:
        Dictionary with parsed BibTeX information

    Raises:
        ValueError: If BibTeX entry is malformed
    """
    bibtex_string = bibtex_string.strip()

    if not bibtex_string.startswith("@"):
        raise ValueError("BibTeX entry must start with '@'")

    # Extract entry type and key
    first_line_match = re.match(r"@(\w+)\s*\{\s*([^,\s]+)", bibtex_string)
    if not first_line_match:
        raise ValueError("Malformed BibTeX entry: cannot parse entry type and key")

    entry_type = first_line_match.group(1).lower()
    key = first_line_match.group(2)

    # Extract fields (simple parsing)
    fields = {}
    field_pattern = r"(\w+)\s*=\s*\{([^}]*)\}"
    for match in re.finditer(field_pattern, bibtex_string):
        field_name = match.group(1).lower()
        field_value = match.group(2)
        fields[field_name] = field_value

    return {"type": entry_type, "key": key, "bibtex": bibtex_string, **fields}


def _load_centralized_bib() -> dict[str, dict[str, str]]:
    """
    Load the internal citations bibliography file.

    Returns:
        Dictionary mapping citation keys to their BibTeX information
    """
    # Find the internal_citations.bib file in the same directory
    current_file = Path(__file__)
    citations_file = current_file.parent / "internal_citations.bib"

    if not citations_file.exists():
        return {}

    citations = {}
    try:
        with open(citations_file, encoding="utf-8") as f:
            content = f.read()

        # Split into individual entries
        entries = re.split(r"\n(?=@)", content)
        for entry in entries:
            entry = entry.strip()
            if entry and entry.startswith("@"):
                try:
                    parsed = _parse_bibtex_entry(entry)
                    citations[parsed["key"]] = parsed
                except ValueError:
                    # Skip malformed entries
                    continue
    except (OSError, UnicodeDecodeError):
        # If we can't read the file, return empty dict
        pass

    return citations


def _validate_citation_key(citation_key: str) -> bool:
    """
    Validate that a citation key exists in the internal bibliography.

    Args:
        citation_key: The citation key to validate

    Returns:
        True if key exists, False otherwise
    """
    centralized_bib = _load_centralized_bib()
    return citation_key in centralized_bib


def register_citation(
    citation: str | list[str], condition: str | None = None, message: str | None = None
) -> None:
    """
    Register citation(s) in the global registry.

    Args:
        citation: Either a citation key, full BibTeX entry, or list of them
        condition: Optional condition for when to cite this reference
        message: Optional custom citation message (defaults to "This work used the methods described in Ref.")
    """
    citations_list = citation if isinstance(citation, list) else [citation]

    # Create a unique group key for this registration
    import uuid

    group_id = str(uuid.uuid4())

    citation_keys = []

    for single_citation in citations_list:
        if single_citation.strip().startswith("@"):
            # Full BibTeX entry
            try:
                parsed = _parse_bibtex_entry(single_citation)
                key = parsed["key"]
                citation_keys.append(key)
                _citation_registry[key] = {
                    "bibtex": single_citation,
                    "condition": condition or "",
                    "message": message
                    or "This work used the methods described in Ref.~",
                    "type": "bibtex",
                    "group_id": group_id,
                }
            except ValueError as e:
                raise ValueError(f"Invalid BibTeX entry: {e}")
        else:
            # Citation key - store without validation
            citation_keys.append(single_citation)
            _citation_registry[single_citation] = {
                "key": single_citation,
                "condition": condition or "",
                "message": message or "This work used the methods described in Ref.~",
                "type": "key",
                "group_id": group_id,
            }

    # Store the group information
    if len(citation_keys) > 1:
        for key in citation_keys:
            _citation_registry[key]["group_keys"] = citation_keys


def reference(
    citation: str | list[str], condition: str | None = None, message: str | None = None
):
    """
    Decorator to add citation tracking to functions and classes.

    This decorator registers the citation(s) and returns the original function/class unchanged.

    Args:
        citation: Either a citation key, full BibTeX entry, or list of them
        condition: Optional condition description for when to cite
        message: Optional custom citation message

    Example:
        @reference("netket3:2021")
        def some_function():
            pass

        @reference(["key1", "key2"], condition="when using multiple methods")
        def another_function():
            pass

        @reference("@article{key, ...}", condition="when using momentum",
                   message="This work implemented the algorithm from")
        def third_function():
            pass
    """

    def decorator(obj):
        # Register the citation(s) - this is all we need to do
        register_citation(citation, condition, message)

        # Return the original function/class unchanged
        return obj

    return decorator


def get_all_citations() -> dict[str, dict[str, str]]:
    """
    Get all registered citations.

    Returns:
        Dictionary mapping citation keys to their information
    """
    all_citations = {}
    centralized_bib = _load_centralized_bib()

    for key, citation_info in _citation_registry.items():
        citation_info = citation_info.copy()

        # If it's a key reference, get the BibTeX from centralized file
        if citation_info.get("type") == "key" and key in centralized_bib:
            citation_info["bibtex"] = centralized_bib[key]["bibtex"]
            citation_info.update(centralized_bib[key])
        elif citation_info.get("type") == "bibtex":
            # For BibTeX entries, parse and add the fields
            try:
                parsed = _parse_bibtex_entry(citation_info["bibtex"])
                citation_info.update(parsed)
            except ValueError:
                # If parsing fails, just keep the original info
                pass

        all_citations[key] = citation_info

    return all_citations


def cite(
    output_file: str | None = None, bib: bool = False, from_cli: bool = False
) -> None:
    """
    Display or export all known citations from decorated functions and classes.

    Args:
        output_file: If provided, write BibTeX to this file. Otherwise print to console.
        bib: If True, generate a 'references.bib' file in the current directory.
        from_cli: If True, show CLI-appropriate instructions instead of Python API instructions.
    """

    netket_version_parts = __version__.split(".")[:2]  # Get major.minor
    netket_version = ".".join(netket_version_parts)

    # Register NetKet's own references as a group if not already present
    netket_papers = ["netket3:2021", "netket2:2019"]
    if not any(paper_key in _citation_registry for paper_key in netket_papers):
        register_citation(
            netket_papers, message=f"This work used NetKet {netket_version}"
        )

    all_citations = get_all_citations()
    if not all_citations:
        console = Console()
        console.print("[yellow]No citations registered.[/yellow]")
        return

    # Handle bib file generation
    bib_filename = output_file or ("references.bib" if bib else None)
    if bib_filename:
        # Write BibTeX file
        with open(bib_filename, "w", encoding="utf-8") as f:
            # Write all citations
            for key, info in all_citations.items():
                if "bibtex" in info:
                    f.write(info["bibtex"] + "\n\n")

        console = Console()
        console.print(
            f"[green]✓[/green] BibTeX file saved to [bold]{bib_filename}[/bold]"
        )

    # Display formatted citation instructions
    console = Console()

    # Group citations by their group_id to handle multiple refs in single registration
    citation_groups = {}

    for key, info in all_citations.items():
        group_id = info.get("group_id", key)  # Use key as fallback group_id
        if group_id not in citation_groups:
            citation_groups[group_id] = {
                "keys": [],
                "condition": info.get("condition", "").strip(),
                "message": info.get(
                    "message", "This work used the methods described in Ref.~"
                ),
            }
        citation_groups[group_id]["keys"].append(key)

    # Separate groups into those with and without conditions
    unconditional_citations = []
    conditional_citations = []

    for group_id, group_info in citation_groups.items():
        keys = group_info["keys"]
        condition = group_info["condition"]
        message = group_info["message"]

        # Create citation with multiple keys if needed
        citation_keys = ",".join(keys)
        citation_text = f"{message}~\\cite{{{citation_keys}}}."

        if not condition:
            # No condition - add to unconditional list
            unconditional_citations.append(citation_text)
        else:
            # Has condition - add to conditional list
            condition_comment = f"%% * {condition}"
            conditional_citations.append(f"{condition_comment}\n% {citation_text}")

    # Build LaTeX content starting with unconditional citations
    latex_content_parts = []

    if unconditional_citations:
        latex_content_parts.append("\n\n".join(unconditional_citations))

    if conditional_citations:
        latex_content_parts.append("\n".join(conditional_citations))

    latex_content = "\n\n".join(latex_content_parts)

    # Create Rich formatted output with improved colors
    title = Text("Code and Algorithmic Acknowledgements", style="bold bright_magenta")

    # Only show instruction if we're displaying citations (not writing to file)
    instruction = Text()
    if from_cli:
        instruction.append("💡 Call with ", style="bright_green")
        instruction.append("netket-cite --bib", style="bold bright_yellow")
        instruction.append(
            " to generate the accompanying bib file in the current directory",
            style="bright_green",
        )
    else:
        instruction.append("💡 Call with ", style="bright_green")
        instruction.append(".cite(bib=True)", style="bold bright_yellow")
        instruction.append(
            " to generate the accompanying bib file in the current directory",
            style="bright_green",
        )

    # Display everything
    console.print()
    console.print(title, justify="center")
    console.print()
    console.print(
        "If you use NetKet for your research, we ask that you acknowledge all relevant publications in your publication.",
        style="bright_cyan",
    )
    console.print(
        "This includes both the NetKet library itself, as well as any specific algorithm and implementation you use.",
        style="bright_cyan",
    )
    console.print(
        "See more information at https://www.netket.org/cite/ ",
        style="bright_cyan",
    )
    console.print()
    console.print(instruction)
    console.print()

    # Print the LaTeX content with indentation (no box for easy copy-paste)
    for line in latex_content.split("\n"):
        if line.strip():  # Only indent non-empty lines
            console.print(f"    {line}")
        else:
            console.print()  # Empty line

    console.print()


def validate_references() -> bool:
    """
    Validate all registered references against the internal bibliography.

    Returns:
        True if all references are valid, False otherwise
    """
    centralized_bib = _load_centralized_bib()
    all_valid = True

    for key, info in _citation_registry.items():
        if info.get("type") == "key":
            if key not in centralized_bib:
                print(f"ERROR: Citation key '{key}' not found in internal bibliography")
                all_valid = False
        elif info.get("type") == "bibtex":
            try:
                _parse_bibtex_entry(info["bibtex"])
            except ValueError as e:
                print(f"ERROR: Invalid BibTeX for key '{key}': {e}")
                all_valid = False

    return all_valid



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps

import jax
from jax import numpy as jnp

from jax.nn import celu as _deprecated_celu
from jax.nn import elu as _deprecated_elu
from jax.nn import gelu as _deprecated_gelu
from jax.nn import glu as _deprecated_glu
from jax.nn import leaky_relu as _deprecated_leaky_relu

from jax.nn import log_sigmoid as _deprecated_log_sigmoid
from jax.nn import log_softmax as _deprecated_log_softmax
from jax.nn import standardize as _deprecated_standardize
from jax.nn import relu as _deprecated_relu

from jax.nn import sigmoid as _deprecated_sigmoid
from jax.nn import soft_sign as _deprecated_soft_sign
from jax.nn import softmax as _deprecated_softmax

from jax.nn import softplus as _deprecated_softplus
from jax.nn import swish as _deprecated_swish
from jax.nn import silu as _deprecated_silu
from jax.nn import selu as _deprecated_selu
from jax.nn import hard_tanh as _deprecated_hard_tanh
from jax.nn import relu6 as _deprecated_relu6
from jax.nn import hard_sigmoid as _deprecated_hard_sigmoid
from jax.nn import hard_swish as _deprecated_hard_swish

from jax.numpy import tanh as _deprecated_tanh
from jax.numpy import cosh as _deprecated_cosh
from jax.numpy import sinh as _deprecated_sinh


from netket.jax import HashablePartial

from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr


def reim(f):
    r"""Modifies a non-linearity to act separately on the real and imaginary parts"""

    @wraps(f)
    def reim_f(f, x):
        sqrt2 = jnp.sqrt(jnp.array(2, dtype=x.real.dtype))
        if jnp.iscomplexobj(x):
            return jax.lax.complex(f(sqrt2 * x.real), f(sqrt2 * x.imag)) / sqrt2
        else:
            return f(x)

    fun = HashablePartial(reim_f, f)

    fun.__name__ = f"reim_{f.__name__}"  # type: ignore
    fun.__doc__ = (
        f"{f.__name__} applied separately to the real and"
        f"imaginary parts of it's input.\n\n"
        f"The docstring to the original function follows.\n\n"
        f"{f.__doc__}"
    )

    return fun


def log_cosh(x):
    """
    Logarithm of the hyperbolic cosine, implemented in a more stable way.
    """
    sgn_x = -2 * jnp.signbit(x.real) + 1
    x = x * sgn_x
    return x + jnp.log1p(jnp.exp(-2.0 * x)) - jnp.log(2.0)


def log_sinh(x):
    """
    Logarithm of the hyperbolic sine.
    """
    return jnp.log(jnp.sinh(x))


def log_tanh(x):
    """
    Logarithm of the hyperbolic tangent.
    """
    return jnp.log(jnp.tanh(x))


reim_selu = reim(_deprecated_selu)
r"""Returns the selu non-linearity, applied separately to the real and imaginary parts"""

reim_relu = reim(_deprecated_relu)
r"""Returns the relu non-linearity, applied separately to the real and imaginary parts"""


# Deprecation machinery for re-exported activation functions
_deprecations = {
    # September 2025, NetKet 3.20
    "celu": (
        "netket.nn.activation.celu is deprecated: use jax.nn.celu directly",
        _deprecated_celu,
    ),
    "elu": (
        "netket.nn.activation.elu is deprecated: use jax.nn.elu directly",
        _deprecated_elu,
    ),
    "gelu": (
        "netket.nn.activation.gelu is deprecated: use jax.nn.gelu directly",
        _deprecated_gelu,
    ),
    "glu": (
        "netket.nn.activation.glu is deprecated: use jax.nn.glu directly",
        _deprecated_glu,
    ),
    "leaky_relu": (
        "netket.nn.activation.leaky_relu is deprecated: use jax.nn.leaky_relu directly",
        _deprecated_leaky_relu,
    ),
    "log_sigmoid": (
        "netket.nn.activation.log_sigmoid is deprecated: use jax.nn.log_sigmoid directly",
        _deprecated_log_sigmoid,
    ),
    "log_softmax": (
        "netket.nn.activation.log_softmax is deprecated: use jax.nn.log_softmax directly",
        _deprecated_log_softmax,
    ),
    "standardize": (
        "netket.nn.activation.standardize is deprecated: use jax.nn.standardize directly",
        _deprecated_standardize,
    ),
    "relu": (
        "netket.nn.activation.relu is deprecated: use jax.nn.relu directly",
        _deprecated_relu,
    ),
    "sigmoid": (
        "netket.nn.activation.sigmoid is deprecated: use jax.nn.sigmoid directly",
        _deprecated_sigmoid,
    ),
    "soft_sign": (
        "netket.nn.activation.soft_sign is deprecated: use jax.nn.soft_sign directly",
        _deprecated_soft_sign,
    ),
    "softmax": (
        "netket.nn.activation.softmax is deprecated: use jax.nn.softmax directly",
        _deprecated_softmax,
    ),
    "softplus": (
        "netket.nn.activation.softplus is deprecated: use jax.nn.softplus directly",
        _deprecated_softplus,
    ),
    "swish": (
        "netket.nn.activation.swish is deprecated: use jax.nn.swish directly",
        _deprecated_swish,
    ),
    "silu": (
        "netket.nn.activation.silu is deprecated: use jax.nn.silu directly",
        _deprecated_silu,
    ),
    "selu": (
        "netket.nn.activation.selu is deprecated: use jax.nn.selu directly",
        _deprecated_selu,
    ),
    "hard_tanh": (
        "netket.nn.activation.hard_tanh is deprecated: use jax.nn.hard_tanh directly",
        _deprecated_hard_tanh,
    ),
    "relu6": (
        "netket.nn.activation.relu6 is deprecated: use jax.nn.relu6 directly",
        _deprecated_relu6,
    ),
    "hard_sigmoid": (
        "netket.nn.activation.hard_sigmoid is deprecated: use jax.nn.hard_sigmoid directly",
        _deprecated_hard_sigmoid,
    ),
    "hard_swish": (
        "netket.nn.activation.hard_swish is deprecated: use jax.nn.hard_swish directly",
        _deprecated_hard_swish,
    ),
    "tanh": (
        "netket.nn.activation.tanh is deprecated: use jax.numpy.tanh directly",
        _deprecated_tanh,
    ),
    "cosh": (
        "netket.nn.activation.cosh is deprecated: use jax.numpy.cosh directly",
        _deprecated_cosh,
    ),
    "sinh": (
        "netket.nn.activation.sinh is deprecated: use jax.numpy.sinh directly",
        _deprecated_sinh,
    ),
}


__getattr__ = _deprecation_getattr(__name__, _deprecations)

del _deprecation_getattr



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Any
from collections.abc import Callable

import jax
import jax.numpy as jnp

from flax import serialization

from netket import jax as nkjax
from netket.hilbert import SpinOrbitalFermions
from netket.models import Slater2nd
from netket.sampler import Sampler
from netket.utils import (
    model_frameworks,
    wrap_to_support_scalar,
    _serialization as serialization_utils,
)
from netket.utils.types import PyTree, SeedT, DType

from netket.vqs.base import VariationalState
from netket.vqs.mc import MCState
from netket.vqs.full_summ import FullSumState
from netket import nn as nknn


@partial(jax.jit, static_argnums=0)
def jit_evaluate(fun: Callable, *args):
    """
    call `fun(*args)` inside of a `jax.jit` frame.

    Args:
        fun: the hashable callable to be evaluated.
        args: the arguments to the function.
    """
    return fun(*args)


@partial(jax.jit, static_argnames=("model", "method"))
def jit_apply(model, *args, method):
    return model.apply(*args, method=method)


class DeterminantVariationalState(VariationalState):
    r"""Variational State for fermionic mean-field states (Hartree-Fock ansatz).

    This state represents a Slater determinant wavefunction using a
    :class:`~netket.models.Slater2nd` model and computes expectation values
    analytically using Wick's theorem rather than Monte Carlo sampling, which is
    efficient.

    The state internally constructs and uses a :class:`~netket.models.Slater2nd` model,
    which supports three types of Hartree-Fock ansätze:

    - **Generalized Hartree-Fock (GHF)**: ``generalized=True``
      Uses :math:`n_{\mathrm{M}} \times n_{\mathrm{f}}` parameters where
      :math:`n_{\mathrm{M}}` is the number of fermionic modes and :math:`n_{\mathrm{f}}`
      is the number of fermions. This is the most flexible ansatz but has the most parameters.

    - **Unrestricted Hartree-Fock (UHF)**: ``generalized=False, restricted=False``
      Uses :math:`n_{\mathrm{S}} \times n_{\mathrm{L}} \times n_{\textrm{f, s}}` parameters where
      :math:`n_{\mathrm{S}}` is the number of spin states, :math:`n_{\mathrm{L}}` is the number
      of spatial orbitals, and :math:`n_{\textrm{f, s}}` is the number of fermions per spin.
      Allows different spatial orbitals for different spins (can describe magnetic states).

    - **Restricted Hartree-Fock (RHF)**: ``generalized=False, restricted=True``
      Uses :math:`n_{\mathrm{L}} \times n_{\textrm{f, s}}` parameters.
      All spins share the same spatial orbitals (appropriate for non-magnetic ground states).

    See :class:`~netket.models.Slater2nd` for more details on the different Hartree-Fock types
    and their mathematical formulation.

    """

    _model_framework: model_frameworks.ModuleFramework | None = None
    """The model framework used to define the model."""

    _model: Any
    """The linen-compatible model definition."""

    _init_fun: Callable | None = None
    """The function used to initialise the parameters and model_state"""

    _apply_fun: Callable
    """The function used to evaluate the model"""

    _rdm: jax.Array | None = None
    """Cached one-body reduced density matrix"""

    def __init__(
        self,
        hilbert: SpinOrbitalFermions,
        *,
        generalized: bool = False,
        restricted: bool = True,
        param_dtype: DType = float,
        variables: PyTree | None = None,
        seed: SeedT | None = None,
    ):
        r"""
        Constructs the DeterminantVariationalState.

        This constructor automatically creates an internal :class:`~netket.models.Slater2nd`
        model with the specified Hartree-Fock type.

        .. warning::

            If unspecified the parameters will be completely random, which is
            a very bad initialization for a Slater Determinant. Consider
            initializing using something more reasonable, like the k-space
            tight-binding orbitals or something similar.

        Args:
            hilbert: The Hilbert space. Must be a :class:`~netket.hilbert.SpinOrbitalFermions`
                with fixed particle number (``n_fermions`` must not be None).
            generalized: If True, uses Generalized Hartree-Fock (GHF) with
                :math:`n_{\mathrm{M}} \times n_{\mathrm{f}}` parameters.
                If False, the type depends on the ``restricted`` parameter.
                See :class:`~netket.models.Slater2nd` for details. Default: False.
            restricted: Only used when ``generalized=False``. If True, uses Restricted
                Hartree-Fock (RHF) with :math:`n_{\mathrm{L}} \times n_{\textrm{f, s}}`
                parameters where all spins share spatial orbitals. If False, uses
                Unrestricted Hartree-Fock (UHF) with separate orbitals per spin.
                See :class:`~netket.models.Slater2nd` for details. Default: True.
            param_dtype: The dtype of the variational parameters (float or complex).
                Default: float.
            variables: Optional dictionary for the initial values for the variables
                (parameters and model state) of the model. If not provided, parameters
                are randomly initialized using the seed.
            seed: Random seed used to generate initial parameters (only used if
                ``variables`` is not provided). Defaults to a random seed.

        Example:
            >>> import netket as nk
            >>> # Create a spin-1/2 fermionic Hilbert space
            >>> hi = nk.hilbert.SpinOrbitalFermions(4, s=1/2, n_fermions_per_spin=(2, 2))
            >>> # Create a Restricted Hartree-Fock state (non-magnetic)
            >>> vstate = nk.experimental.vqs.DeterminantVariationalState(
            ...     hi, generalized=False, restricted=True, seed=42
            ... )
            >>> print(vstate.n_parameters)  # Number of variational parameters
        """
        # Validate hilbert space
        if not isinstance(hilbert, SpinOrbitalFermions):
            raise TypeError(
                "DeterminantVariationalState requires a SpinOrbitalFermions hilbert space, "
                f"but got {type(hilbert)}"
            )

        if hilbert.n_fermions is None:
            raise ValueError(
                "DeterminantVariationalState requires a hilbert space with a fixed number "
                "of fermions (n_fermions must not be None)."
            )

        super().__init__(hilbert)

        # Extract init and apply functions
        model = Slater2nd(
            hilbert,
            generalized=generalized,
            restricted=restricted,
            param_dtype=param_dtype,
        )
        self._model_framework = model_frameworks.identify_framework(model)
        _maybe_unwrapped_variables, model = self._model_framework.wrap(model)

        if variables is None:
            if _maybe_unwrapped_variables is not None:
                variables = _maybe_unwrapped_variables

        self._model = model

        self._init_fun = nkjax.HashablePartial(
            lambda model, *args, **kwargs: model.init(*args, **kwargs), model
        )
        self._apply_fun = wrap_to_support_scalar(
            nkjax.HashablePartial(
                lambda model, *args, **kwargs: model.apply(*args, **kwargs), model
            )
        )

        if variables is not None:
            self.variables = variables
        else:
            self.init(seed, dtype=float)

        self._rdm = None
        self.mutable = False

    @property
    def model(self) -> Any | None:
        """Returns the model definition of this variational state.

        This field is optional, and is set to `None` if the variational state has
        been initialized using a custom function.
        """
        if self._model_framework is not None:
            return self._model_framework.unwrap(self._model, self.variables)
        return self._model

    def init(self, seed=None, dtype=None):
        """
        Initialises the variational parameters of the variational state.
        """
        if self._init_fun is None:
            raise RuntimeError(
                "Cannot initialise the parameters of this state "
                "because you did not supply a valid init_function."
            )

        if dtype is None:
            dtype = float

        key = nkjax.PRNGKey(seed)
        dummy_input = self.hilbert.random_state(key, 1, dtype=dtype)

        variables = jit_evaluate(self._init_fun, {"params": key}, dummy_input)
        self.variables = variables

    def reset(self):
        """
        Resets the cached density matrix. Called automatically when
        parameters/state is updated.
        """
        self._rdm = None

    def log_value(self, σ: jnp.ndarray) -> jnp.ndarray:
        return jit_evaluate(self._apply_fun, self.variables, σ)

    @property
    def rdm(self) -> jax.Array:
        """
        The one-body reduced density matrix of the mean-field state.

        This is computed from the Slater determinant parameters and cached
        until parameters are updated.

        Returns:
            The density matrix of shape (n_modes, n_modes) where n_modes is
            hilbert.size.
        """
        if self._rdm is None:
            self._rdm = jit_apply(
                self._model,
                self.variables,
                method=self._model.rdm_one_body,
            )
        return self._rdm  # type: ignore

    def quantum_geometric_tensor(self, qgt_T=None):
        r"""
        Compute the centered quantum geometric tensor (QGT) for the Slater determinant.

        For a Slater determinant, the QGT can be computed analytically using the formula:

        .. math::

            Q_{(\mu i),(\nu j)} = \delta_{ij} \left[\delta_{\mu\nu} - \gamma_{\mu\nu}\right]

        where :math:`\gamma = C C^\dagger` is the one-body reduced density matrix.

        This method returns a :class:`~netket.optimizer.linear_operator.DenseOperator`
        wrapping the dense QGT matrix, compatible with the standard NetKet QGT interface.

        Args:
            qgt_T: Optional QGT type/constructor. If None, defaults to QGTByWick.

        Returns:
            A :class:`~netket.optimizer.linear_operator.DenseOperator` wrapping
            the centered QGT matrix of shape (n_parameters, n_parameters).

        Note:
            - **Generalized HF**: Matrix of shape (n_modes × n_fermions, n_modes × n_fermions)
            - **Restricted HF**: Matrix of shape (n_orbitals × n_fermions_per_spin, n_orbitals × n_fermions_per_spin)
            - **Unrestricted HF**: Block-diagonal matrix with blocks for each spin sector

        Example:
            >>> hi = nk.hilbert.SpinOrbitalFermions(4, s=1/2, n_fermions_per_spin=(2, 2))
            >>> vstate = nk.experimental.vqs.DeterminantVariationalState(
            ...     hi, generalized=False, restricted=True, seed=42
            ... )
            >>> qgt = vstate.quantum_geometric_tensor(diag_shift=0.01)
            >>> print(qgt.matrix.shape)  # (8, 8)
            >>> print(vstate.n_parameters)  # 8
        """
        if qgt_T is None:
            from netket._src.vqs.fermion_mf.qgt import QGTByWick

            qgt_T = QGTByWick

        return qgt_T(self)

    def to_fullsumstate(self, **kwargs):
        """
        Convert this DeterminantVariationalState to a FullSumState.

        Warning:
            Only feasible for small systems due to exponential Hilbert space size.

        Args:
            **kwargs: Additional arguments passed to FullSumState constructor
                     (e.g., chunk_size)

        Returns:
            A FullSumState with the same model and parameters
        """
        return FullSumState(
            hilbert=self.hilbert,
            model=self.model,
            variables=self.variables,
            **kwargs,
        )

    def to_mcstate(
        self,
        sampler: Sampler,
        n_samples: int | None = None,
        n_samples_per_rank: int | None = None,
        **kwargs,
    ):
        """
        Convert this DeterminantVariationalState to an MCState for sampling-based calculations.

        Args:
            sampler: The sampler to use for the MCState (e.g., MetropolisFermionHop)
            n_samples: Number of samples for Monte Carlo estimation
            n_samples_per_rank: Alternative specification of samples per MPI rank
            **kwargs: Additional arguments passed to MCState constructor
                     (e.g., n_discard_per_chain, chunk_size)

        Returns:
            An MCState with the same model and parameters

        Example:
            >>> mf_state = DeterminantVariationalState(hilbert, model)
            >>> sampler = nk.sampler.MetropolisFermionHop(hilbert, graph=g)
            >>> mc_state = mf_state.to_mcstate(sampler, n_samples=1024)
            >>> qgt = mc_state.quantum_geometric_tensor()
        """
        if sampler.hilbert != self.hilbert:
            raise ValueError(
                f"Sampler hilbert space {sampler.hilbert} does not match "
                f"state hilbert space {self.hilbert}"
            )

        return MCState(
            sampler=sampler,
            model=self.model,
            n_samples=n_samples,
            n_samples_per_rank=n_samples_per_rank,
            variables=self.variables,
            **kwargs,
        )

    def to_array(self, normalize: bool = True) -> jax.Array:
        return nknn.to_array(
            self.hilbert,
            self._apply_fun,
            self.variables,
            normalize=normalize,
        )

    def __repr__(self):
        return (
            "DeterminantVariationalState("
            + f"\n  hilbert = {self.hilbert},"
            + f"\n  generalized = {self.model.generalized},"
            + f"\n  restricted = {self.model.restricted},"
            + f"\n  n_parameters = {self.n_parameters})"
        )


# Serialization support


def serialize_DeterminantVariationalState(vstate):
    state_dict = {
        "variables": serialization.to_state_dict(
            serialization_utils.remove_prngkeys(vstate.variables)
        ),
    }
    return state_dict


def deserialize_DeterminantVariationalState(vstate, state_dict):
    import copy

    new_vstate = copy.copy(vstate)
    new_vstate.reset()

    vars = jax.tree_util.tree_map(
        jnp.asarray,
        serialization.from_state_dict(vstate.variables, state_dict["variables"]),
    )
    vars = serialization_utils.restore_prngkeys(vstate.variables, vars)

    new_vstate.variables = vars
    return new_vstate


serialization.register_serialization_state(
    DeterminantVariationalState,
    serialize_DeterminantVariationalState,
    deserialize_DeterminantVariationalState,
)



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Citation system for NetKet.

This module provides tools for citation tracking and bibliography generation.
"""

__all__ = [
    "cite",
    "reference",
    "get_all_citations",
    "validate_references",
    "get_citations_for_docs",
]


from netket.utils.citations.core import (
    cite,
    reference,
    get_all_citations,
    validate_references,
)
from netket.utils.citations.docs import (
    get_citations_for_docs,
)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Ignore false-positives for redefined `product` functions:
# pylint: disable=function-redefined

import itertools

import numpy as np
from typing import overload, Literal, Any

import jax
import jax.numpy as jnp

from netket.utils import HashableArray, struct
from netket.utils.types import Array, DType, Shape
from netket.utils.dispatch import dispatch

from netket.utils import warn_deprecation, deprecated, deprecated_new_name

from netket.utils.group._group import FiniteGroup
from netket.utils.group._semigroup import Element


class Permutation(Element):
    def __init__(
        self,
        permutation: Array | None = None,
        *,
        name: str | None = None,
        permutation_array: Array | None = None,
        inverse_permutation_array: Array | None = None,
        validate: bool = True,
    ):
        r"""
        Creates a `Permutation` from either the array of images
        `permutation_array` or preimages `inverse_permutation_array`.

        Exactly one argument among `permutation_array` and
        `inverse_permutation_array` (and the deprecated argument `permutation`)
        must be specified.

        The deprecated argument `permutation` should be substituted for
        `inverse_permutation_array`.

        Note that the left action of a permutation on an array `a` is
        `a[inverse_permutation_array]`.

        Args:
            permutation: (deprecated) 1D array listing
                :math:`g^{-1}(x)` for all :math:`0\le x \le N-1`.
            name: Optional, custom name for the permutation.
            permutation_array: 1D array listing
                :math:`g(x)` for all :math:`0\le x \le N-1`.
            inverse_permutation_array: 1D array listing
                :math:`g^{-1}(x)` for all :math:`0\le x \le N-1`.
            validate: If true (Default) validates the array of indices
                passed to permutation

        Returns:
            A `Permutation` object that encodes the specified permutation.
        """

        arg_list = [permutation, permutation_array, inverse_permutation_array]
        if sum([arg is not None for arg in arg_list]) != 1:
            raise ValueError(
                "Exactly one argument among `permutation`, `permutation_array` "
                "and `inverse_permutation_array` must be specified."
            )

        if permutation is not None:
            warn_deprecation(
                "The argument `permutation` is deprecated.\n\n"
                "In order to clarify notations, you should either pass "
                "the array of images `permutation_array` or "
                "preimages `inverse_permutation_array`."
            )
            inverse_permutation_array = permutation

        if validate:
            if permutation_array is not None:
                permutation_elements = sorted(np.array(permutation_array))
            else:
                permutation_elements = sorted(np.array(inverse_permutation_array))

            els = np.asarray(permutation_elements)
            n = len(els)

            # Check if it's a valid permutation: all elements in [0, n-1] and no duplicates
            # Fast validation: sort and compare with expected range
            if n == 0 or not np.array_equal(np.sort(els), np.arange(n)):
                raise ValueError(
                    "The indices of the permutation are invalid. "
                    "A permutation over n elements should be specified by an array with "
                    "elements in {0, 1, ..., n-1}."
                )

        if permutation_array is not None:
            inverse_permutation_array = np.argsort(permutation_array)

        self._inverse_permutation_array = HashableArray(
            np.asarray(inverse_permutation_array)
        )

        self.__name = name

    def __hash__(self):
        return hash(self._inverse_permutation_array)

    def __eq__(self, other):
        if isinstance(other, Permutation):
            return np.array_equal(self.permutation_array, other.permutation_array)
        else:
            return False

    @property
    def permutation_array(self):
        return np.asarray(np.argsort(self._inverse_permutation_array))

    @property
    def inverse_permutation_array(self):
        return np.asarray(self._inverse_permutation_array)

    @property
    @deprecated(
        "Deprecated in favor of `permutation.inverse_permutation_array` or "
        "`permutation.permutation_array`"
    )
    def permutation(self):
        return self._inverse_permutation_array

    @property
    def _name(self):
        return self.__name

    def __repr__(self):
        if self._name is not None:
            return self._name
        else:
            return f"Permutation({self.permutation_array.tolist()})"

    def cycle_decomposition(self) -> list:
        """
        Return the cycle decomposition of the permutation.

        The returned object is a list of each disjoint cycle of the permutation as a list.
        """
        permutation_array = self.permutation_array
        cycle_list = []
        visited = np.zeros(len(permutation_array), dtype=bool)
        while not np.all(visited):
            starting_point = np.nonzero(1 - visited)[0][0]
            current_point = starting_point
            cycle_list.append([])
            while not visited[starting_point]:
                current_point = permutation_array[current_point]
                cycle_list[-1].append(current_point.item())
                visited[current_point] = True
        return cycle_list

    @deprecated_new_name("permutation.inverse_permutation_array")
    def __array__(self, dtype: DType = None):
        return np.asarray(self._inverse_permutation_array, dtype)

    def apply_to_id(self, x: Array):
        """Returns the image of indices `x` under the permutation"""
        return self.permutation_array[x]


@dispatch
def product(p: Permutation, x: Array):
    # if p.permutation is a HashableArray and x is a jax Array
    # direct indexing fails, so we call np.asarray on it to extract the
    # wrapped array
    # TODO make indexing work with HashableArray directly

    if isinstance(x, jax.Array):
        return x.at[..., p.inverse_permutation_array].get(
            unique_indices=True, mode="promise_in_bounds"
        )
    else:
        return x[..., p.inverse_permutation_array]


@dispatch
def product(p: Permutation, q: Permutation):  # noqa: F811
    name = None if p._name is None and q._name is None else f"{p} @ {q}"
    inverse_permutation_array = q.inverse_permutation_array[p.inverse_permutation_array]
    return Permutation(
        inverse_permutation_array=inverse_permutation_array, name=name, validate=False
    )


@struct.dataclass
class PermutationGroup(FiniteGroup):
    r"""
    Collection of permutation operations acting on sequences of length :code:`degree`.

    Group elements need not all be of type :class:`netket.utils.group.Permutation`,
    only act as such on a sequence when called.

    The class can contain elements that are distinct as objects (e.g.,
    :code:`Identity()` and :code:`Translation((0,))`) but have identical action.
    Those can be removed by calling
    :meth:`~netket.utils.group.PermutationGroup.remove_duplicates`.
    """

    degree: int
    """Number of elements the permutations act on."""

    def __hash__(self):
        return super().__hash__()

    def __eq__(self, other: "PermutationGroup") -> bool:
        if isinstance(other, PermutationGroup):
            return np.array_equal(self.to_array(), other.to_array())
        return False

    def _canonical(self, x: Element) -> Array:
        return x(np.arange(self.degree, dtype=int))

    def to_array(self) -> Array:
        r"""
        Convert the abstract group operations to an array of permutation indices.


        It returns a matrix where the `i`-th row contains the indices corresponding
        to the `i`-th group element. That is, :code:`self.to_array()[i, j]`
        is :math:`g_i^{-1}(j)`. Moreover,

        .. code::

            G = # this permutation group...
            V = np.arange(G.degree)
            assert np.all(G(V) == V[..., G.to_array()])

        Returns:
            A matrix that can be used to index arrays in the computational basis
            in order to obtain their permutations.
        """
        return self._canonical_array()

    def __array__(self, dtype=None) -> Array:
        return np.asarray(self.to_array(), dtype=dtype)

    @overload
    def remove_duplicates(
        self, *, return_inverse: Literal[False] = False
    ) -> "PermutationGroup": ...

    @overload
    def remove_duplicates(
        self, *, return_inverse: Literal[True]
    ) -> tuple["PermutationGroup", Any]: ...

    def remove_duplicates(self, *, return_inverse=False):
        r"""
        Returns a new :code:`PermutationGroup` with duplicate elements (that is,
        elements which represent identical permutations) removed.

        Args:
            return_inverse: If `True`, also return indices to reconstruct the original
                group from the result.

        Returns:
            The permutation group with duplicate elements removed. If
            :code:`return_inverse==True`, it also returns the indices needed to
            reconstruct the original group from the result.
        """
        if return_inverse:
            group, inverse = super().remove_duplicates(return_inverse=True)
        else:
            group = super().remove_duplicates(return_inverse=False)

        pgroup = PermutationGroup(group.elems, self.degree)

        if return_inverse:
            return pgroup, inverse
        else:
            return pgroup

    @struct.property_cached
    def inverse(self) -> Array:
        try:
            lookup = self._canonical_lookup()
            inverses = []
            # `np.argsort` on a 1D permutation list generates the inverse permutation
            # it acts along last axis by default, so can perform it on to_array()
            # `np.argsort` changes int32 to int64 on Windows,
            # and we need to change it back
            perms = self.to_array()
            invperms = np.argsort(perms).astype(perms.dtype)

            for invperm in invperms:
                inverses.append(lookup[HashableArray(invperm)])

            return np.asarray(inverses, dtype=int)
        except KeyError as err:
            raise RuntimeError(
                "PermutationGroup does not contain the inverse of all elements"
            ) from err

    @struct.property_cached
    def product_table(self) -> Array:
        perms = self.to_array()
        inverse = perms[self.inverse].squeeze()
        n_symm = len(perms)
        lookup = np.unique(np.column_stack((perms, np.arange(len(self)))), axis=0)

        product_table = np.zeros([n_symm, n_symm], dtype=int)
        for i, g_inv in enumerate(inverse):
            row_perms = perms[:, g_inv]
            row_perms = np.unique(
                np.column_stack((row_perms, np.arange(len(self)))), axis=0
            )
            # row_perms should be a permutation of perms, so identical after sorting
            if np.any(row_perms[:, :-1] != lookup[:, :-1]):
                raise RuntimeError(
                    "PermutationGroup is not closed under multiplication"
                )
            # match elements in row_perms to group indices
            product_table[i, row_perms[:, -1]] = lookup[:, -1]

        return product_table

    @property
    def shape(self) -> Shape:
        r"""
        Tuple :code:`(<# of group elements>, <degree>)`.

        Equivalent to :code:`self.to_array().shape`.
        """
        return (len(self), self.degree)

    def apply_to_id(self, x: Array):
        """Returns the image of indices `x` under all permutations"""
        return self.to_array()[self.inverse][:, x]


PermutationGroup.inverse.__doc__ = FiniteGroup.inverse.__doc__
PermutationGroup.product_table.__doc__ = FiniteGroup.product_table.__doc__


@dispatch
def product(A: PermutationGroup, B: PermutationGroup):  # noqa: F811
    if A.degree != B.degree:
        raise ValueError(
            "Incompatible groups (`PermutationGroup`s of different degree)"
        )
    return PermutationGroup(
        elems=[a @ b for a, b in itertools.product(A.elems, B.elems)], degree=A.degree
    )


@dispatch
def product(G: PermutationGroup, x: Array):  # noqa: F811

    if isinstance(x, jax.Array):
        return jnp.moveaxis(
            x.at[..., G.to_array()].get(unique_indices=True, mode="promise_in_bounds"),
            -2,
            0,
        )
    else:
        return np.moveaxis(x[..., G.to_array()], -2, 0)



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

import jax.numpy as jnp
import jax

from netket.vqs import MCState, expect
from netket.stats import statistics
from netket import jax as nkjax

from .S2_operator import Renyi2EntanglementEntropy


@expect.dispatch
def Renyi2(vstate: MCState, op: Renyi2EntanglementEntropy, chunk_size: int | None):
    if op.hilbert != vstate.hilbert:
        raise TypeError("Hilbert spaces should match")

    samples = vstate.samples
    n_chains = samples.shape[0]
    n_samples = samples.shape[0] * samples.shape[1]

    if n_chains % 2 != 0 and not vstate.sampler.is_exact:
        raise ValueError("Use an even number of chains.")

    if n_chains == 1:
        if n_samples % 2 != 0:
            samples = samples[:, :-1]
        σ_η = samples[:, : (n_samples // 2)]
        σp_ηp = samples[:, (n_samples // 2) :]

    else:
        σ_η = samples[: (n_chains // 2)]
        σp_ηp = samples[(n_chains // 2) :]

    return Renyi2_sampling_MCState(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        σ_η,
        σp_ηp,
        op.partition,
        chunk_size=chunk_size,
    )


@partial(jax.jit, static_argnames=("afun", "chunk_size"))
def Renyi2_sampling_MCState(
    afun, params, model_state, σ_η, σp_ηp, partition, *, chunk_size
):
    n_chains = σ_η.shape[0]

    N = σ_η.shape[-1]

    σ_η = σ_η.reshape(-1, N)
    σp_ηp = σp_ηp.reshape(-1, N)

    n_samples = σ_η.shape[0]

    σ = σ_η[:, partition]
    σp = σp_ηp[:, partition]

    σ_ηp = jnp.copy(σp_ηp)
    σp_η = jnp.copy(σ_η)

    σ_ηp = σ_ηp.at[:, partition].set(σ)
    σp_η = σp_η.at[:, partition].set(σp)

    @partial(
        nkjax.apply_chunked, in_axes=(None, None, 0, 0, 0, 0), chunk_size=chunk_size
    )
    def kernel_fun(params, model_state, σ_ηp, σp_η, σ_η, σp_ηp):
        W = {"params": params, **model_state}

        return jnp.exp(afun(W, σ_ηp) + afun(W, σp_η) - afun(W, σ_η) - afun(W, σp_ηp))

    kernel_values = kernel_fun(params, model_state, σ_ηp, σp_η, σ_η, σp_ηp)

    Renyi2_stats = statistics(kernel_values.reshape((n_chains, -1)))

    # Propagation of errors from S_2 to -log_2(S_2)
    Renyi2_stats = Renyi2_stats.replace(
        variance=Renyi2_stats.variance / (Renyi2_stats.mean.real * jnp.log(2)) ** 2
    )

    Renyi2_stats = Renyi2_stats.replace(
        error_of_mean=jnp.sqrt(Renyi2_stats.variance / n_samples)
    )

    Renyi2_stats = Renyi2_stats.replace(mean=-jnp.log2(Renyi2_stats.mean).real)

    return Renyi2_stats



def safe_map(f, *args):
    args = list(map(list, args))
    n = len(args[0])
    for arg in args[1:]:
        assert len(arg) == n, f"length mismatch: {list(map(len, args))}"
    return list(map(f, *args))



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import jax
import numpy as np
from jax import numpy as jnp

from netket.utils import config, struct

from . import mean as _mean
from . import var as _var
from . import total_size as _total_size
from ._autocorr import integrated_time


def _format_main_string(value, std, var):

    if not math.isfinite(abs(value)) or not math.isfinite(std):
        return f"{value:.2e} ± {std:.2e} [σ²={var:.1e}"

    elif std == 0.0:
        return f"{value:.3e} [σ²={var:.1e}"

    elif (abs(value) + abs(std) < 1e-2) or abs(std) <= 1e-7:
        value_std_str = _format_scientific_notation(value, std)
        return value_std_str + f" [σ²={var:.1e}"

    else:
        if std < 1e-15:
            decimals = 15
        else:
            decimals = max(int(np.ceil(-np.log10(std))), 0) + 1
        return f"{value:.{decimals}f} ± {std:.{decimals}f} [σ²={var:.1e}"


def _format_scientific_notation(value, std):

    length = 5

    value_ref = (
        np.real(value)
        if np.abs(np.real(value)) > np.abs(np.imag(value))
        else np.imag(value)
    )

    exponent_val = int(np.floor(np.log10(np.abs(value_ref))))
    exponent_std = int(np.floor(np.log10(std)))
    n_digits = max(exponent_val - exponent_std, 0) + 1

    mantissa_val = value / 10**exponent_val
    mantissa_std = std / 10**exponent_val

    value_length = 2 * length + 2 if isinstance(value, complex) else length

    mantissa_str = (
        f"{mantissa_val:.{n_digits}f}".rjust(value_length)
        + " ± "
        + f"{mantissa_std:.{n_digits}f}".rjust(length)
    )

    return mantissa_str + " e" + f"{value_ref:e}".split("e")[1]


_NaN = float("NaN")


def _maybe_item(x):
    if hasattr(x, "shape") and x.shape == ():
        return x.item()
    else:
        return x


@struct.dataclass
class Stats:
    """A dict-compatible pytree containing the result of the statistics function."""

    mean: float | complex = _NaN
    """The mean value."""
    error_of_mean: float = _NaN
    """Estimate of the error of the mean."""
    variance: float = _NaN
    """Estimation of the variance of the data."""
    tau_corr: float = _NaN
    """Estimate of the autocorrelation time (in dimensionless units of number of steps).

    This value is estimated with a blocking algorithm by default, but the result is known
    to be unreliable. A more precise estimator based on the FFT transform can be used by
    setting the environment variable `NETKET_EXPERIMENTAL_FFT_AUTOCORRELATION=1`. This
    estimator is more computationally expensive, but overall the added cost should be
    negligible.
    """
    R_hat: float = _NaN
    """
    Estimator of the split-Rhat convergence estimator.

    The split-Rhat diagnostic is based on comparing intra-chain and inter-chain
    statistics of the sample and is thus only available for 2d-array inputs where
    the rows are independently sampled MCMC chains. In an ideal MCMC samples,
    R_hat should be 1.0. If it deviates from this value too much, this indicates
    MCMC convergence issues. Thresholds such as R_hat > 1.1 or even R_hat > 1.01 have
    been suggested in the literature for when to discard a sample. (See, e.g.,
    Gelman et al., `Bayesian Data Analysis <http://www.stat.columbia.edu/~gelman/book/>`_,
    or Vehtari et al., `arXiv:1903.08008 <https://arxiv.org/abs/1903.08008>`_.)
    """
    tau_corr_max: float = _NaN
    """
    Estimate of the maximum autocorrelation time among all Markov chains.

    This value is only computed if the environment variable
    `NETKET_EXPERIMENTAL_FFT_AUTOCORRELATION` is set.
    """

    def to_dict(self):
        jsd = {}
        jsd["Mean"] = _maybe_item(self.mean)
        jsd["Variance"] = _maybe_item(self.variance)
        jsd["Sigma"] = _maybe_item(self.error_of_mean)
        jsd["R_hat"] = _maybe_item(self.R_hat)
        jsd["TauCorr"] = _maybe_item(self.tau_corr)
        if config.netket_experimental_fft_autocorrelation:
            jsd["TauCorrMax"] = _maybe_item(self.tau_corr_max)
        return jsd

    def to_compound(self):
        return "Mean", self.to_dict()

    def __repr__(self):
        main_string = _format_main_string(
            _maybe_item(self.mean),
            _maybe_item(self.error_of_mean),
            _maybe_item(self.variance),
        )
        if not math.isnan(self.R_hat):
            ext = f", R̂={self.R_hat:.3f}"
        else:
            ext = ""
        if config.netket_experimental_fft_autocorrelation:
            if not (math.isnan(self.tau_corr) and math.isnan(self.tau_corr_max)):
                ext += f", τ={self.tau_corr:.1f}<{self.tau_corr_max:.1f}"
        return main_string + f"{ext}]"

    # Alias accessors
    def __getattr__(self, name):
        if name in ("mean", "Mean"):
            return self.mean
        elif name in ("variance", "Variance"):
            return self.variance
        elif name in ("error_of_mean", "Sigma"):
            return self.error_of_mean
        elif name in ("R_hat", "R"):
            return self.R_hat
        elif name in ("tau_corr", "TauCorr"):
            return self.tau_corr
        elif name in ("tau_corr_max", "TauCorrMax"):
            return self.tau_corr_max
        else:
            raise AttributeError(f"'Stats' object object has no attribute '{name}'")

    def real(self):
        return self.replace(mean=np.real(self.mean))

    def imag(self):
        return self.replace(mean=np.imag(self.mean))


def _get_blocks(data, block_size):
    chain_length = data.shape[1]

    n_blocks = int(np.floor(chain_length / float(block_size)))

    return data[:, 0 : n_blocks * block_size].reshape((-1, block_size)).mean(axis=1)


def _block_variance(data, l):
    blocks = _get_blocks(data, l)
    ts = _total_size(blocks)
    if ts > 0:
        return _var(blocks), ts
    else:
        return jnp.nan, 0


def _batch_variance(data):
    b_means = data.mean(axis=1)
    ts = _total_size(b_means)
    return _var(b_means), ts


def _split_R_hat(data, W):
    N = data.shape[-1]
    if not config.netket_use_plain_rhat:
        # compute split-chain batch variance
        local_batch_size = data.shape[0]
        if N % 2 == 0:
            # split each chain in the middle,
            # like [[1 2 3 4]] -> [[1 2][3 4]]
            batch_var, _ = _batch_variance(data.reshape(2 * local_batch_size, N // 2))
        else:
            # drop the last sample of each chain for an even split,
            # like [[1 2 3 4 5]] -> [[1 2][3 4]]
            batch_var, _ = _batch_variance(
                data[:, :-1].reshape(2 * local_batch_size, N // 2)
            )

    # V_loc = _np.var(data, axis=-1, ddof=0)
    # W_loc = _np.mean(V_loc)
    # W = _mean(W_loc)
    # # This approximation seems to hold well enough for larger n_samples
    return jnp.sqrt((N - 1) / N + batch_var / W)


BLOCK_SIZE = 32


def statistics(data: jax.Array) -> Stats:
    r"""
    Returns statistics of a given array (or matrix, see below) containing a stream of data.
    This is particularly useful to analyze Markov Chain data, but it can be used
    also for other type of time series.

    Args:
        data (vector or matrix): The input data. It can be real or complex valued.
            * if a vector, it is assumed that this is a time series of data (not necessarily independent);
            * if a matrix, it is assumed that that rows :code:`data[i]` contain independent time series.

    Returns:
       Stats:
        A dictionary-compatible class containing the
        average (:code:`.mean`, :code:`["Mean"]`),
        variance (:code:`.variance`, :code:`["Variance"]`),
        the Monte Carlo standard error of the mean (:code:`error_of_mean`, :code:`["Sigma"]`),
        an estimate of the autocorrelation time (:code:`tau_corr`, :code:`["TauCorr"]`), and the
        Gelman-Rubin split-Rhat diagnostic (:code:`.R_hat`, :code:`["R_hat"]`).

        If the flag `NETKET_EXPERIMENTAL_FFT_AUTOCORRELATION` is set, the autocorrelation is computed
        exactly using a FFT transform, and an extra field `tau_corr_max` is inserted in the
        statistics object

        These properties can be accessed both the attribute and the dictionary-style syntax
        (both indicated above).

        The split-Rhat diagnostic is based on comparing intra-chain and inter-chain
        statistics of the sample and is thus only available for 2d-array inputs where
        the rows are independently sampled MCMC chains. In an ideal MCMC samples,
        R_hat should be 1.0. If it deviates from this value too much, this indicates
        MCMC convergence issues. Thresholds such as R_hat > 1.1 or even R_hat > 1.01 have
        been suggested in the literature for when to discard a sample. (See, e.g.,
        Gelman et al., `Bayesian Data Analysis <http://www.stat.columbia.edu/~gelman/book/>`_,
        or Vehtari et al., `arXiv:1903.08008 <https://arxiv.org/abs/1903.08008>`_.)
    """
    if config.netket_experimental_fft_autocorrelation:
        return _statistics(data)
    else:
        from .mc_stats_old import statistics as statistics_blocks

        return statistics_blocks(data)


@jax.jit
def _statistics(data):
    data = jnp.atleast_1d(data)
    if data.ndim == 1:
        data = data.reshape((1, -1))

    if data.ndim > 2:
        raise NotImplementedError("Statistics are implemented only for ndim<=2")

    mean = _mean(data)
    variance = _var(data)

    taus = jax.vmap(integrated_time)(data)
    tau_avg = jnp.mean(taus)
    tau_max = jnp.max(taus)

    batch_var, n_batches = _batch_variance(data)
    if n_batches > 1:
        error_of_mean = jnp.sqrt(batch_var / n_batches)
        R_hat = _split_R_hat(data, variance)
    else:
        l_block = max(1, data.shape[1] // BLOCK_SIZE)
        block_var, n_blocks = _block_variance(data, l_block)
        error_of_mean = jnp.sqrt(block_var / n_blocks)
        R_hat = jnp.nan

    res = Stats(mean, error_of_mean, variance, tau_avg, R_hat, tau_max)

    return res



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Import NetKet-specific activation functions for deprecation machinery
from .activation import (
    log_cosh as _deprecated_log_cosh,
    log_sinh as _deprecated_log_sinh,
    log_tanh as _deprecated_log_tanh,
    reim_selu as _deprecated_reim_selu,
    reim_relu as _deprecated_reim_relu,
    reim as _deprecated_reim,
)

# Import deprecated functions for deprecation machinery
from jax.nn import celu as _deprecated_celu
from jax.nn import elu as _deprecated_elu
from jax.nn import gelu as _deprecated_gelu
from jax.nn import glu as _deprecated_glu
from jax.nn import leaky_relu as _deprecated_leaky_relu
from jax.nn import log_sigmoid as _deprecated_log_sigmoid
from jax.nn import log_softmax as _deprecated_log_softmax
from jax.nn import relu as _deprecated_relu
from jax.nn import sigmoid as _deprecated_sigmoid
from jax.nn import soft_sign as _deprecated_soft_sign
from jax.nn import softmax as _deprecated_softmax
from jax.nn import softplus as _deprecated_softplus
from jax.nn import swish as _deprecated_swish
from jax.nn import silu as _deprecated_silu
from jax.numpy import tanh as _deprecated_tanh
from jax.numpy import cosh as _deprecated_cosh
from jax.numpy import sinh as _deprecated_sinh

from .symmetric_linear import (
    DenseSymm,
    DenseEquivariant,
)

from netket.nn.masked_linear import MaskedDense1D, MaskedConv1D, MaskedConv2D
from netket.nn.fast_masked_linear import (
    FastMaskedDense1D,
    FastMaskedConv1D,
    FastMaskedConv2D,
)

from netket.nn.utils import (
    to_array,
    to_matrix,
    binary_encoding,
)

from netket.nn import blocks as blocks
from netket.nn import activation as activation
from netket.nn import apply_operator as apply_operator

# Import for deprecation machinery
from netket._src.nn.apply_operator.functional import (
    make_logpsi_op_afun as _deprecated_make_logpsi_op_afun,
)


# Deprecation machinery for activation functions
_deprecations = {
    # December 2024, NetKet 3.15
    "celu": (
        "netket.nn.celu is deprecated: use jax.nn.celu directly",
        _deprecated_celu,
    ),
    "elu": (
        "netket.nn.elu is deprecated: use jax.nn.elu directly",
        _deprecated_elu,
    ),
    "gelu": (
        "netket.nn.gelu is deprecated: use jax.nn.gelu directly",
        _deprecated_gelu,
    ),
    "glu": (
        "netket.nn.glu is deprecated: use jax.nn.glu directly",
        _deprecated_glu,
    ),
    "leaky_relu": (
        "netket.nn.leaky_relu is deprecated: use jax.nn.leaky_relu directly",
        _deprecated_leaky_relu,
    ),
    "log_sigmoid": (
        "netket.nn.log_sigmoid is deprecated: use jax.nn.log_sigmoid directly",
        _deprecated_log_sigmoid,
    ),
    "log_softmax": (
        "netket.nn.log_softmax is deprecated: use jax.nn.log_softmax directly",
        _deprecated_log_softmax,
    ),
    "relu": (
        "netket.nn.relu is deprecated: use jax.nn.relu directly",
        _deprecated_relu,
    ),
    "sigmoid": (
        "netket.nn.sigmoid is deprecated: use jax.nn.sigmoid directly",
        _deprecated_sigmoid,
    ),
    "soft_sign": (
        "netket.nn.soft_sign is deprecated: use jax.nn.soft_sign directly",
        _deprecated_soft_sign,
    ),
    "softmax": (
        "netket.nn.softmax is deprecated: use jax.nn.softmax directly",
        _deprecated_softmax,
    ),
    "softplus": (
        "netket.nn.softplus is deprecated: use jax.nn.softplus directly",
        _deprecated_softplus,
    ),
    "swish": (
        "netket.nn.swish is deprecated: use jax.nn.swish directly",
        _deprecated_swish,
    ),
    "silu": (
        "netket.nn.silu is deprecated: use jax.nn.silu directly",
        _deprecated_silu,
    ),
    "tanh": (
        "netket.nn.tanh is deprecated: use jax.numpy.tanh directly",
        _deprecated_tanh,
    ),
    "cosh": (
        "netket.nn.cosh is deprecated: use jax.numpy.cosh directly",
        _deprecated_cosh,
    ),
    "sinh": (
        "netket.nn.sinh is deprecated: use jax.numpy.sinh directly",
        _deprecated_sinh,
    ),
    # NetKet-specific activation functions that should point to netket.nn.activation
    "log_cosh": (
        "netket.nn.log_cosh is deprecated: use netket.nn.activation.log_cosh",
        _deprecated_log_cosh,
    ),
    "log_sinh": (
        "netket.nn.log_sinh is deprecated: use netket.nn.activation.log_sinh",
        _deprecated_log_sinh,
    ),
    "log_tanh": (
        "netket.nn.log_tanh is deprecated: use netket.nn.activation.log_tanh",
        _deprecated_log_tanh,
    ),
    "reim_selu": (
        "netket.nn.reim_selu is deprecated: use netket.nn.activation.reim_selu",
        _deprecated_reim_selu,
    ),
    "reim_relu": (
        "netket.nn.reim_relu is deprecated: use netket.nn.activation.reim_relu",
        _deprecated_reim_relu,
    ),
    "reim": (
        "netket.nn.reim is deprecated: use netket.nn.activation.reim",
        _deprecated_reim,
    ),
    # December 2025, NetKet 3.21
    "make_logpsi_op_afun": (
        "netket.nn.make_logpsi_op_afun is deprecated: use netket.nn.apply_operator.make_logpsi_op_afun",
        _deprecated_make_logpsi_op_afun,
    ),
}


from netket.utils.deprecation import deprecation_getattr as _deprecation_getattr

__getattr__ = _deprecation_getattr(__name__, _deprecations)

del _deprecation_getattr



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Icosahedral symmetry groups.

This module provides functions to construct icosahedral point groups,
including both the pure rotational group (I) and the full icosahedral
group with inversion (Ih). These are the symmetry groups of the icosahedron
and dodecahedron.
"""

from netket.utils.moduletools import export, hide_unexported

from netket.utils.group.axial import inversion_group as _inv_group
from netket.utils.group.axial import C as _C
from netket.utils.group._point_group import PointGroup


hide_unexported(__name__)

__all__ = [
    "icosahedral_rotations",
    "icosahedral",
]


@export
def I() -> PointGroup:  # noqa: E743
    """
    Rotational symmetries of an icosahedron with two vertices on z axis
    and two others in the xz plane.
    """
    g1 = _C(5, axis=(0, 0, 1))
    g2 = _C(5, axis=(2 / 5**0.5, 0, 1 / 5**0.5))
    g = (g2 @ g1).remove_duplicates()
    g = (g @ g2).remove_duplicates()
    g = (g @ g1).remove_duplicates()
    return g


icosahedral_rotations = I


@export
def Ih() -> PointGroup:
    """
    Symmetry group of an icosahedron with two vertices on z axis
    and two others in the xz plane.
    """
    return _inv_group() @ I()


icosahedral = Ih



# Copyright 2025 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp
from jax.tree_util import register_pytree_node_class

from netket.hilbert import HomogeneousHilbert
from netket.hilbert.constraint import SumConstraint

from netket._src.operator.permutation.base import (
    PermutationOperatorBase,
)
from netket._src.operator.permutation.trace_utils import count_n_uplets


@register_pytree_node_class
class PermutationOperator(PermutationOperatorBase):
    """
    Permutation operator on a spin or boson space. Used for
    the symmetry-representation machinery.

    For mathematical details on the definition of a permutation operator
    and its justification, we refer to :doc:`/advanced/symmetry`.

    For the fermionic counterpart look at
    :class:`netket.operator.permutation.PermutationOperatorFermion`.
    """

    def get_conn_padded(self, x):
        r"""Find the connected elements of the operator.

        Starting from a batch of quantum numbers :math:`x={x_1, ... x_n}` of
        size :math:`B \times M` where :math:`B` size of the batch and :math:`M`
        size of the hilbert space, finds all states :math:`y_i^1, ..., y_i^K`
        connected to every :math:`x_i`.

        Returns a matrix of size :math:`B \times K_{max} \times M` where
        :math:`K_{max}` is the maximum number of connections for every
        :math:`y_i`.

        .. warning::

            Unlike most other operators defined in NetKet, a permutation operator
            is not Hermitian, and we thus have to be careful about the definition of
            connected elements. NetKet defines connected elements of :math:`x` as the
            configurations :math:`x'` such that :math:`\langle x | P_\sigma | x' \rangle`.
            Therefore, the connected elements are the configurations found in the
            image of :math:`x` by :math:`P_{\sigma^{-1}}`, and not :math:`P_\sigma`.

        Args:
            x : A N-tensor of shape :code:`(...,hilbert.size)` containing
                the batch/batches of quantum numbers :math:`x`.

        Returns:
            **(x_primes, mels)**: The connected states x', in a N+1-tensor and an
            N-tensor containing the matrix elements :math:`O(x,x')`
            associated to each x' for every batch.
        """
        x = jnp.asarray(x)
        # Check that the parameters of get are useful
        x_conn = x.at[..., None, self.permutation.permutation_array].get(
            unique_indices=True, mode="promise_in_bounds"
        )

        # we want to do
        # mels = jnp.ones((*x.shape[:-1], 1), dtype=self.dtype)
        # but to preserve sharding we must do
        mels = x.at[..., :1].get(unique_indices=True, mode="promise_in_bounds")
        mels = 1 + (mels * 0).astype(self.dtype)

        return x_conn, mels

    def trace(self) -> int:
        cycle_decomposition = self.permutation.cycle_decomposition()

        if isinstance(self.hilbert, HomogeneousHilbert):

            if not self.hilbert.constrained:
                return self.hilbert.local_size ** len(cycle_decomposition)

            elif isinstance(self.hilbert.constraint, SumConstraint):
                cycle_lengths = [len(cycle) for cycle in cycle_decomposition]
                shifted_constraint = (
                    self.hilbert.constraint.sum_value
                    + self.hilbert.size * (self.hilbert.local_size - 1)
                )
                return count_n_uplets(
                    cycle_lengths, shifted_constraint, self.hilbert.local_size
                )

            else:
                raise NotImplementedError(
                    f"Unimplemented trace for Hilbert constraint {type(self.hilbert.constraint)}"
                )

        else:
            raise NotImplementedError(
                f"Unimplemented trace for Hilbert space {type(self.hilbert).__name__}"
            )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
import warnings

import jax
from jax import numpy as jnp
from jax.flatten_util import ravel_pytree
from flax import struct

import netket.jax as nkjax
from netket.utils import timing, HashablePartial
from netket.utils.types import PyTree
from netket.utils.api_utils import partial_from_kwargs
from netket.errors import (
    IllegalHolomorphicDeclarationForRealParametersError,
    NonHolomorphicQGTOnTheFlyDenseRepresentationError,
    HolomorphicUndeclaredWarning,
)

from .common import check_valid_vector_type
from .qgt_onthefly_logic import mat_vec_factory, mat_vec_chunked_factory

from ..linear_operator import LinearOperator, SolverT, Uninitialized


@partial_from_kwargs
def QGTOnTheFly(
    vstate, *, chunk_size=None, holomorphic: bool | None = None, **kwargs
) -> "QGTOnTheFlyT":
    """
    Lazy representation of an S Matrix computed by performing 2 jvp
    and 1 vjp products, using the variational state's model, the
    samples that have already been computed, and the vector.

    The S matrix is not computed yet, but can be computed by calling
    :code:`to_dense`.
    The details on how the ⟨S⟩⁻¹⟨F⟩ system is solved are contained in
    the field `sr`.

    Args:
        vstate: The variational State.
        chunk_size: If supplied, overrides the chunk size of the variational state
                    (useful for models where the backward pass requires more
                    memory than the forward pass).
    """
    if kwargs.pop("diag_scale", None) is not None:
        raise NotImplementedError(
            "\n`diag_scale` argument is not yet supported by QGTOnTheFly."
            "Please use `QGTJacobianPyTree` or `QGTJacobianDense`.\n\n"
            "You are also encouraged to nag the developers to support "
            "this feature.\n\n"
        )

    # TODO: Find a better way to handle this case
    from netket.vqs import FullSumState

    if isinstance(vstate, FullSumState):
        samples = vstate._all_states
        pdf = vstate.probability_distribution()
    else:
        samples = vstate.samples
        pdf = None

    if chunk_size is None:
        chunk_size = getattr(vstate, "chunk_size", None)

    return QGTOnTheFly_DefaultConstructor(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        samples,
        pdf=pdf,
        chunk_size=chunk_size,
        holomorphic=holomorphic,
        **kwargs,
    )


@timing.timed
def QGTOnTheFly_DefaultConstructor(
    apply_fun,
    parameters,
    model_state,
    samples,
    pdf=None,
    *,
    chunk_size: int | None = None,
    holomorphic: bool | None = None,
    **kwargs,
) -> "QGTOnTheFlyT":
    """ """
    if pdf is not None:
        if not pdf.shape == samples.shape[:-1]:
            raise ValueError(
                "The shape of pdf must match the shape of the samples, "
                f"instead you provided (pdf.shape={pdf.shape}) != "
                f"(samples.shape={samples.shape[:-1]})"
            )
        if pdf.ndim >= 2:
            pdf = jax.jit(jax.lax.collapse, static_argnums=(1, 2))(pdf, 0, 2)

    # The code does not support an extra batch dimension
    if samples.ndim >= 3:
        # use jit so that we can do it on global shared array
        samples = jax.jit(jax.lax.collapse, static_argnums=(1, 2))(samples, 0, 2)

    n_samples_per_rank = samples.shape[0] // jax.device_count()
    if chunk_size is None or chunk_size >= n_samples_per_rank:
        mv_factory = mat_vec_factory
        chunking = False
    else:
        mv_factory = HashablePartial(mat_vec_chunked_factory, chunk_size=chunk_size)
        chunking = True

    # check if holomorphic or not
    if holomorphic:
        if nkjax.tree_leaf_isreal(parameters):
            raise IllegalHolomorphicDeclarationForRealParametersError()
        else:
            mode = "holomorphic"
    else:
        if not nkjax.tree_leaf_iscomplex(parameters):
            mode = "real"
        else:
            if holomorphic is None:
                warnings.warn(HolomorphicUndeclaredWarning(), UserWarning)
            mode = "complex"

    nkjax.jacobian_default_mode(
        apply_fun,
        parameters,
        model_state,
        samples,
        holomorphic=holomorphic,
    )

    mat_vec = mv_factory(
        forward_fn=apply_fun,
        params=parameters,
        model_state=model_state,
        samples=samples,
        pdf=pdf,
    )
    return QGTOnTheFlyT(
        _mat_vec=mat_vec,
        _params=parameters,
        _chunking=chunking,
        _mode=mode,
        **kwargs,
    )


@struct.dataclass
class QGTOnTheFlyT(LinearOperator):
    """
    Lazy representation of an S Matrix computed by performing 2 jvp
    and 1 vjp products, using the variational state's model, the
    samples that have already been computed, and the vector.

    The S matrix is not computed yet, but can be computed by calling
    :code:`to_dense`.
    The details on how the ⟨S⟩⁻¹⟨F⟩ system is solved are contained in
    the field `sr`.
    """

    _mat_vec: Callable[[PyTree, float], PyTree] = Uninitialized  # type: ignore
    """The S matrix-vector product as generated by mat_vec_factory.
    It's a jax.Partial, so can be used as pytree_node."""

    _params: PyTree = Uninitialized  # type: ignore
    """The first input to apply_fun (parameters of the ansatz).
    Only used as a shape placeholder."""

    _chunking: bool = struct.field(pytree_node=False, default=False)
    """Whether the implementation with chunks is used which currently does not support vmapping over it"""

    _mode: str = struct.field(pytree_node=False, default=None)
    """Differentiation mode:
        - "real": for real-valued R->R and C->R Ansätze, splits the complex inputs
                  into real and imaginary part.
        - "complex": for complex-valued R->C and C->C Ansätze, splits the complex
                  inputs and outputs into real and imaginary part
        - "holomorphic": for any Ansätze. Does not split complex values.
        - "auto": autoselect real or complex.
    """

    def __matmul__(self, y):
        return onthefly_mat_treevec(self, y)

    def _solve(
        self, solve_fun: SolverT, y: PyTree, *, x0: PyTree | None, **kwargs
    ) -> PyTree:
        return _solve(self, solve_fun, y, x0=x0)

    def to_dense(self) -> jnp.ndarray:
        """
        Convert the lazy matrix representation to a dense matrix representation.

        Returns:
            A dense matrix representation of this S matrix.
        """
        # This condition will be true if the user specified `holomorphic=False` and
        # if the parameters are complex. If the parameters are real and the user
        # did not specify holomorphic we will have mode==real and if holomorphic is
        # True mode==holomorphic.
        #
        # We must check this because the AD implementation will compute the wrong
        # QGT in that case
        if self._mode == "complex":
            raise NonHolomorphicQGTOnTheFlyDenseRepresentationError()

        return _to_dense(self)

    def __repr__(self):
        return f"QGTOnTheFly(diag_shift={self.diag_shift})"


@jax.jit
def onthefly_mat_treevec(
    S: QGTOnTheFly, vec: PyTree | jnp.ndarray
) -> PyTree | jnp.ndarray:
    """
    Perform the lazy mat-vec product, where vec is either a tree with the same structure as
    params or a ravelled vector
    """

    # if has a ndim it's an array and not a pytree
    if hasattr(vec, "ndim"):
        if not vec.ndim == 1:
            raise ValueError("Unsupported mat-vec for chunks of vectors")
        # If the input is a vector
        if not nkjax.tree_size(S._params) == vec.size:
            raise ValueError(
                """Size mismatch between number of parameters ({nkjax.tree_size(S.params)})
                                and vector size {vec.size}.
                             """
            )

        _, unravel = ravel_pytree(S._params)
        vec = unravel(vec)
        ravel_result = True
    else:
        ravel_result = False

    check_valid_vector_type(S._params, vec)

    vec = nkjax.tree_cast(vec, S._params)

    res = S._mat_vec(vec, S.diag_shift)

    if ravel_result:
        res, _ = ravel_pytree(res)

    return res


@jax.jit
def _solve(
    self: QGTOnTheFlyT, solve_fun, y: PyTree, *, x0: PyTree | None, **kwargs
) -> PyTree:
    check_valid_vector_type(self._params, y)

    y = nkjax.tree_cast(y, self._params)

    # we could cache this...
    if x0 is None:
        x0 = jax.tree_util.tree_map(jnp.zeros_like, y)

    out, info = solve_fun(self, y, x0=x0)
    return out, info


@jax.jit
def _to_dense(self: QGTOnTheFlyT) -> jnp.ndarray:
    """
    Convert the lazy matrix representation to a dense matrix representation

    Returns:
        A dense matrix representation of this S matrix.
    """
    Npars = nkjax.tree_size(self._params)
    I = jax.numpy.eye(Npars)

    if self._chunking:
        # the linear_call in mat_vec_chunked does currently not have a jax batching rule,
        # so it cannot be vmapped but we can use scan
        # which is better for reducing the memory consumption anyway
        _, out = jax.lax.scan(lambda _, x: (None, self @ x), None, I)
    else:
        out = jax.vmap(lambda x: self @ x, in_axes=0)(I)

    if jnp.iscomplexobj(out):
        out = out.T

    return out



# Copyright 2021-2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

import numpy as np
import numba

from netket.errors import concrete_or_error, NumbaOperatorGetConnDuringTracingError


from .compile_helpers import pack_internals
from .base import LocalOperatorBase

if TYPE_CHECKING:
    from .jax import LocalOperatorJax


class LocalOperatorNumba(LocalOperatorBase):
    """Numba implementation of an operator composed of a sum of
    local terms, each of which acts on a small number of sites.

    .. warning::

        The complexity of the constructor of this operator scales
        exponentially with the number of sites on which each term
        acts, and linearly with the number of terms.

        Refrain from using this for terms acting on more than
        6-sites. For Spin-1/2 systems, prefer instead
        :class:`netket.operator.PauliStrings`. For non spin-1/2
        there is nothing that will work efficiently out of the box,
        but you can easily roll your own.

    .. warning::

        The runtime complexity is proportional to the number of
        connected entries
        :attr:`~netket.operator.DiscreteOperator.max_conn_size`.

    For the jax-based implementation, look at
    :class:`netket.operator.LocalOperatorJax`.
    """

    __module__ = "netket.operator"

    def _setup(self, force: bool = False):
        """Analyze the operator strings and precompute arrays for get_conn inference"""
        if force or not self._initialized:
            data = pack_internals(
                self.hilbert,
                self._operators_dict,
                self.constant,
                self.dtype,
                self.mel_cutoff,
            )

            self._acting_on = data["acting_on"]
            self._acting_size = data["acting_size"]
            self._diag_mels = data["diag_mels"]
            self._mels = data["mels"]
            self._x_prime = data["x_prime"]
            self._n_conns = data["n_conns"]
            self._basis = data["basis"]
            self._nonzero_diagonal = data["nonzero_diagonal"]
            self._max_conn_size = data["max_conn_size"]
            self._initialized = True

    def get_conn_flattened(self, x, sections, pad=False):
        r"""Finds the connected elements of the Operator. Starting
        from a given quantum number x, it finds all other quantum numbers x' such
        that the matrix element :math:`O(x,x')` is different from zero. In general there
        will be several different connected states x' satisfying this
        condition, and they are denoted here :math:`x'(k)`, for :math:`k=0,1...N_{\mathrm{connected}}`.

        This is a batched version, where x is a matrix of shape (batch_size,hilbert.size).

        Args:
            x (matrix): A matrix of shape (batch_size,hilbert.size) containing
                        the batch of quantum numbers x.
            sections (array): An array of size (batch_size) useful to unflatten
                        the output of this function.
                        See numpy.split for the meaning of sections.
            pad (bool): Whether to use zero-valued matrix elements in order to return all equal sections.

        Returns:
            matrix: The connected states x', flattened together in a single matrix.
            array: An array containing the matrix elements :math:`O(x,x')` associated to each x'.

        """
        self._setup()

        x_ids = self.hilbert.states_to_local_indices(x)

        x_ids = concrete_or_error(
            np.asarray,
            x_ids,
            NumbaOperatorGetConnDuringTracingError,
            self,
        )

        xp_ids, mels = self._get_conn_flattened_kernel(
            x_ids,
            sections,
            self._basis,
            self._constant,
            self._diag_mels,
            self._n_conns,
            self._mels,
            self._x_prime,
            self._acting_on,
            self._acting_size,
            self._nonzero_diagonal,
            pad,
        )

        xp = self.hilbert.local_indices_to_states(xp_ids, dtype=x.dtype)
        return xp, mels

    @staticmethod
    @numba.jit(nopython=True)
    def _get_conn_flattened_kernel(
        x,
        sections,
        basis,
        constant,
        diag_mels,
        n_conns,
        all_mels,
        all_x_prime,
        acting_on,
        acting_size,
        nonzero_diagonal,
        pad=False,
    ):
        batch_size = x.shape[0]
        n_sites = x.shape[1]
        dtype = all_mels.dtype

        # TODO remove this line when numba 0.53 is dropped 0.54 is minimum version
        # workaround a bug in Numba arising when NUMBA_BOUNDSCHECK=1
        constant = constant.item()

        assert sections.shape[0] == batch_size

        n_operators = n_conns.shape[0]
        # array to store the row index
        xs_n = np.empty((batch_size, n_operators), dtype=np.intp)

        tot_conn = 0
        max_conn = 0

        for b in range(batch_size):
            # diagonal element
            conn_b = 1 if nonzero_diagonal else 0

            # counting the off-diagonal elements
            for i in range(n_operators):
                acting_size_i = acting_size[i]
                # compute the number (row index) from the local states
                # (here we do the inverse of _number_to_state from
                # compile_helpers.py, so this is essentially _state_to_number)
                xs_n[b, i] = 0
                x_b = x[b]
                x_i = x_b[acting_on[i, :acting_size_i]]
                # iterate over sites the current operator is acting on
                for k in range(acting_size_i):
                    # compute
                    xs_n[b, i] += x_i[acting_size_i - k - 1] * basis[i, k]

                # sum the number of off-diagonal connected elements
                conn_b += n_conns[i, xs_n[b, i]]

            tot_conn += conn_b
            sections[b] = tot_conn

            if pad:
                max_conn = max(conn_b, max_conn)

        if pad:
            tot_conn = batch_size * max_conn

        x_prime = np.empty((tot_conn, n_sites), dtype=x.dtype)
        mels = np.empty(tot_conn, dtype=dtype)

        c = 0
        for b in range(batch_size):
            c_diag = c
            x_batch = x[b]

            if nonzero_diagonal:
                mels[c_diag] = constant
                x_prime[c_diag] = np.copy(x_batch)
                c += 1

            for i in range(n_operators):
                if nonzero_diagonal:
                    mels[c_diag] += diag_mels[i, xs_n[b, i]]

                # get the number of connected elements for the current operator
                # at the rows index corresponding to the state of x at the sites
                # the operator is acting on
                n_conn_i = n_conns[i, xs_n[b, i]]

                if n_conn_i > 0:
                    sites = acting_on[i]
                    acting_size_i = acting_size[i]

                    for cc in range(n_conn_i):  # iterate over compressed nonzero cols
                        # get the nonzero mels of the current row
                        mels[c + cc] = all_mels[i, xs_n[b, i], cc]
                        x_prime[c + cc] = np.copy(x_batch)
                        # set the changed local states of the sites the operator
                        # is acting on
                        # it is stored in all_x_prime, where we select the row
                        for k in range(acting_size_i):
                            x_prime[c + cc, sites[k]] = all_x_prime[
                                i, xs_n[b, i], cc, k
                            ]
                    c += n_conn_i

            if pad:
                delta_conn = max_conn - (c - c_diag)
                mels[c : c + delta_conn].fill(0)
                x_prime[c : c + delta_conn, :] = np.copy(x_batch)
                c += delta_conn
                sections[b] = c

        return x_prime, mels

    def get_conn_filtered(self, x, sections, filters):
        r"""Finds the connected elements of the Operator using only a subset of operators. Starting
        from a given quantum number x, it finds all other quantum numbers x' such
        that the matrix element :math:`O(x,x')` is different from zero. In general there
        will be several different connected states x' satisfying this
        condition, and they are denoted here :math:`x'(k)`, for :math:`k=0,1...N_{\mathrm{connected}}`.

        This is a batched version, where x is a matrix of shape (batch_size,hilbert.size).

        Args:
            x (matrix): A matrix of shape (batch_size,hilbert.size) containing
                        the batch of quantum numbers x.
            sections (array): An array of size (batch_size) useful to unflatten
                        the output of this function.
                        See numpy.split for the meaning of sections.
            filters (array): Only operators op(filters[i]) are used to find the connected elements of
                        x[i].

        Returns:
            matrix: The connected states x', flattened together in a single matrix.
            array: An array containing the matrix elements :math:`O(x,x')` associated to each x'.

        """
        self._setup()

        x_ids = self.hilbert.states_to_local_indices(x)
        x_ids = concrete_or_error(
            np.asarray,
            x_ids,
            NumbaOperatorGetConnDuringTracingError,
            self,
        )

        xp_ids, mels = self._get_conn_filtered_kernel(
            x_ids,
            sections,
            self._basis,
            self._constant,
            self._diag_mels,
            self._n_conns,
            self._mels,
            self._x_prime,
            self._acting_on,
            self._acting_size,
            filters,
        )
        xp = self.hilbert.local_indices_to_states(xp_ids, dtype=x.dtype)
        return xp, mels

    @staticmethod
    @numba.jit(nopython=True)
    def _get_conn_filtered_kernel(
        x,
        sections,
        basis,
        constant,
        diag_mels,
        n_conns,
        all_mels,
        all_x_prime,
        acting_on,
        acting_size,
        filters,
    ):
        batch_size = x.shape[0]
        n_sites = x.shape[1]
        dtype = all_mels.dtype

        assert filters.shape[0] == batch_size and sections.shape[0] == batch_size

        # TODO remove this line when numba 0.53 is dropped 0.54 is minimum version
        # workaround a bug in Numba arising when NUMBA_BOUNDSCHECK=1
        constant = constant.item()

        n_operators = n_conns.shape[0]
        xs_n = np.empty((batch_size, n_operators), dtype=np.intp)

        tot_conn = 0

        for b in range(batch_size):
            # diagonal element
            tot_conn += 1

            # counting the off-diagonal elements
            i = filters[b]

            assert i < n_operators and i >= 0
            acting_size_i = acting_size[i]

            xs_n[b, i] = 0
            x_b = x[b]
            x_i = x_b[acting_on[i, :acting_size_i]]
            for k in range(acting_size_i):
                xs_n[b, i] += x_i[acting_size_i - k - 1] * basis[i, k]

            tot_conn += n_conns[i, xs_n[b, i]]
            sections[b] = tot_conn

        x_prime = np.empty((tot_conn, n_sites))
        mels = np.empty(tot_conn, dtype=dtype)

        c = 0
        for b in range(batch_size):
            c_diag = c
            mels[c_diag] = constant
            x_batch = x[b]
            x_prime[c_diag] = np.copy(x_batch)
            c += 1

            i = filters[b]
            # Diagonal part
            mels[c_diag] += diag_mels[i, xs_n[b, i]]
            n_conn_i = n_conns[i, xs_n[b, i]]

            if n_conn_i > 0:
                sites = acting_on[i]
                acting_size_i = acting_size[i]

                for cc in range(n_conn_i):
                    mels[c + cc] = all_mels[i, xs_n[b, i], cc]
                    x_prime[c + cc] = np.copy(x_batch)

                    for k in range(acting_size_i):
                        x_prime[c + cc, sites[k]] = all_x_prime[i, xs_n[b, i], cc, k]
                c += n_conn_i

        return x_prime, mels

    def to_jax_operator(self) -> "LocalOperatorJax":  # noqa: F821
        """
        Returns the jax-compatible version of this operator, which is an
        instance of :class:`netket.operator.LocalOperatorJax`.
        """
        from .jax import LocalOperatorJax

        return LocalOperatorJax(
            self.hilbert,
            self.operators,
            self.acting_on,
            self.constant,
            dtype=self.dtype,
            mel_cutoff=self.mel_cutoff,
        )


# Keep the old name for backward compatibility
LocalOperator = LocalOperatorNumba
